import binascii
from ..errors import EncodeError as _EncodeError
from ..errors import DecodeError as _DecodeError


class EncodeError(_EncodeError):
    """General ASN.1 encode error.

    """

    pass


class DecodeError(_DecodeError):
    """General ASN.1 decode error.

    """

    def __init__(self, message):
        super(DecodeError, self).__init__()
        self.message = message
        self.location = []

    def __str__(self):
        return "{}: {}".format(': '.join(self.location[::-1]),
                               self.message)


class DecodeTagError(DecodeError):
    """ASN.1 tag decode error.

    """

    def __init__(self, type_name, expected_tag, actual_tag, offset):
        message = "expected {} with tag '{}' at offset {}, but got '{}'".format(
            type_name,
            binascii.hexlify(expected_tag).decode('ascii'),
            offset,
            binascii.hexlify(actual_tag).decode('ascii'))
        super(DecodeTagError, self).__init__(message)


class DecodeContentsLengthError(DecodeError):
    """ASN.1 contents length decode error.

    """

    def __init__(self, length, offset, contents_max):
        message = ('expected at least {} contents byte(s) at offset {}, '
                   'but got {}').format(length,
                                        offset,
                                        contents_max - offset)
        super(DecodeContentsLengthError, self).__init__(message)

        self.length = length
        self.offset = offset
        self.contents_max = contents_max



import unittest
from binascii import hexlify


def format_encoded(encoded):
    hexstring = hexlify(encoded).decode('ascii')
    binstring = bin(int('80' + hexstring, 16))[10:]
    binstring = ' '.join([binstring[i:i + 8]
                          for i in range(0, len(binstring), 8)])

    return '{} ({})'.format(hexstring, binstring)


class Asn1ToolsBaseTest(unittest.TestCase):

    def assert_encode_decode(self,
                             specification,
                             type_name,
                             decoded_message,
                             encoded_message):
        try:
            encoded = specification.encode(type_name, decoded_message)
        except:
            print('Failed to encode type with name {}.'.format(type_name))
            raise

        try:
            self.assertEqual(encoded, encoded_message)
        except:
            print('Wrong encoding of type with name {}.'.format(type_name))
            print('Actual:', format_encoded(encoded))
            print('Expected:', format_encoded(encoded_message))
            raise

        try:
            decoded = specification.decode(type_name, encoded_message)
        except:
            print('Failed to decode type with name {}.'.format(type_name))
            raise

        try:
            self.assertEqual(decoded, decoded_message)
        except:
            print('Wrong decoding of type with name {}.'.format(type_name))
            print('Actual:', decoded)
            print('Expected:', decoded_message)
            raise



"""XML Encoding Rules (XER) codec.

"""

import logging
from xml.etree import ElementTree
import binascii

from . import EncodeError
from . import DecodeError
from . import compiler
from .compiler import enum_values_as_dict


LOGGER = logging.getLogger(__name__)


class Type(object):

    def __init__(self, name, type_name):
        self.name = name
        self.type_name = type_name
        self.optional = False
        self.default = None

    def encode(self, data):
        raise NotImplementedError()

    def encode_of(self, data):
        return self.encode(data)

    def decode(self, element):
        raise NotImplementedError()

    def decode_of(self, element):
        return self.decode(element)


class StringType(Type):

    def encode(self, data):
        element = ElementTree.Element(self.name)

        if len(data) > 0:
            element.text = data

        return element

    def decode(self, element):
        if element.text is None:
            return u''
        else:
            return element.text

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__,
                               self.name)


class Integer(Type):

    def __init__(self, name):
        super(Integer, self).__init__(name, 'INTEGER')

    def encode(self, data):
        element = ElementTree.Element(self.name)
        element.text = str(data)

        return element

    def decode(self, element):
        return int(element.text)

    def __repr__(self):
        return 'Integer({})'.format(self.name)


class Real(Type):

    def __init__(self, name):
        super(Real, self).__init__(name, 'REAL')

    def encode(self, data):
        raise NotImplementedError()

    def decode(self, element):
        raise NotImplementedError()

    def __repr__(self):
        return 'Real({})'.format(self.name)


class Boolean(Type):

    def __init__(self, name):
        super(Boolean, self).__init__(name, 'BOOLEAN')

    def encode(self, data):
        element = ElementTree.Element(self.name)
        ElementTree.SubElement(element, 'true' if data else 'false')

        return element

    def decode(self, element):
        return element.find('true') is not None

    def encode_of(self, data):
        return ElementTree.Element('true' if data else 'false')

    def decode_of(self, element):
        return element.tag == 'true'

    def __repr__(self):
        return 'Boolean({})'.format(self.name)


class MembersType(Type):

    def __init__(self, name, members, type_name):
        super(MembersType, self).__init__(name, type_name)
        self.members = members

    def encode(self, data):
        element = ElementTree.Element(self.name)

        for member in self.members:
            name = member.name

            if name in data:
                member_element = member.encode(data[name])
            elif member.optional or member.default is not None:
                continue
            else:
                raise EncodeError(
                    "{} member '{}' not found in {}.".format(
                        self.__class__.__name__,
                        name,
                        data))

            element.append(member_element)

        return element

    def decode(self, element):
        values = {}

        for member in self.members:
            name = member.name
            member_element = element.find(name)

            if member_element is not None:
                value = member.decode(member_element)
                values[name] = value
            elif member.optional:
                pass
            elif member.default is not None:
                values[name] = member.default

        return values

    def __repr__(self):
        return '{}({}, [{}])'.format(
            self.__class__.__name__,
            self.name,
            ', '.join([repr(member) for member in self.members]))


class Sequence(MembersType):

    def __init__(self, name, members):
        super(Sequence, self).__init__(name, members, 'SEQUENCE')


class Set(MembersType):

    def __init__(self, name, members):
        super(Set, self).__init__(name, members, 'SET')


class ArrayType(Type):

    def __init__(self, name, element_type, type_name):
        super(ArrayType, self).__init__(name, type_name)
        self.element_type = element_type

    def encode(self, data):
        element = ElementTree.Element(self.name)

        for entry in data:
            element.append(self.element_type.encode_of(entry))

        return element

    def decode(self, element):
        values = []

        for member_element in list(element):
            value = self.element_type.decode_of(member_element)
            values.append(value)

        return values

    def __repr__(self):
        return '{}({}, {})'.format(self.__class__.__name__,
                                   self.name,
                                   self.element_type)


class SequenceOf(ArrayType):

    def __init__(self, name, element_type):
        super(SequenceOf, self).__init__(name,
                                         element_type,
                                         'SEQUENCE OF')


class SetOf(ArrayType):

    def __init__(self, name, element_type):
        super(SetOf, self).__init__(name,
                                    element_type,
                                    'SET OF')


class BitString(Type):

    def __init__(self, name):
        super(BitString, self).__init__(name, 'BIT STRING')

    def encode(self, data):
        element = ElementTree.Element(self.name)

        if data[1] > 0:
            encoded = int(binascii.hexlify(data[0]), 16)
            encoded |= (0x80 << (8 * len(data[0])))
            element.text = bin(encoded)[10:10 + data[1]].upper()

        return element

    def decode(self, element):
        encoded = element.text

        if encoded is None:
            number_of_bits = 0
            decoded = b''
        else:
            number_of_bits = len(encoded)
            decoded = int(encoded, 2)
            decoded |= (0x80 << number_of_bits)
            rest = (number_of_bits % 8)

            if rest != 0:
                decoded <<= (8 - rest)

            decoded = binascii.unhexlify(hex(decoded)[4:])

        return (decoded, number_of_bits)

    def __repr__(self):
        return 'BitString({})'.format(self.name)


class OctetString(Type):

    def __init__(self, name):
        super(OctetString, self).__init__(name, 'OCTET STRING')

    def encode(self, data):
        element = ElementTree.Element(self.name)

        if len(data) > 0:
            element.text = binascii.hexlify(data).decode('ascii').upper()

        return element

    def decode(self, element):
        if element.text is None:
            return b''
        else:
            return binascii.unhexlify(element.text)

    def __repr__(self):
        return 'OctetString({})'.format(self.name)


class IA5String(StringType):

    def __init__(self, name):
        super(IA5String, self).__init__(name, 'IA5String')


class NumericString(StringType):

    def __init__(self, name):
        super(NumericString, self).__init__(name, 'NumericString')


class PrintableString(StringType):

    def __init__(self, name):
        super(PrintableString, self).__init__(name, 'PrintableString')


class UniversalString(StringType):

    def __init__(self, name):
        super(UniversalString, self).__init__(name, 'UniversalString')


class VisibleString(StringType):

    def __init__(self, name):
        super(VisibleString, self).__init__(name, 'VisibleString')


class GeneralString(StringType):

    def __init__(self, name):
        super(GeneralString, self).__init__(name, 'GeneralString')


class UTF8String(StringType):

    def __init__(self, name):
        super(UTF8String, self).__init__(name, 'UTF8String')


class BMPString(StringType):

    def __init__(self, name):
        super(BMPString, self).__init__(name, 'BMPString')


class GraphicString(StringType):

    def __init__(self, name):
        super(GraphicString, self).__init__(name, 'GraphicString')


class UTCTime(Type):

    def __init__(self, name):
        super(UTCTime, self).__init__(name, 'UTCTime')

    def encode(self, data):
        raise NotImplementedError()

    def decode(self, element):
        raise NotImplementedError()

    def __repr__(self):
        return 'UTCTime({})'.format(self.name)


class GeneralizedTime(Type):

    def __init__(self, name):
        super(GeneralizedTime, self).__init__(name, 'GeneralizedTime')

    def encode(self, data):
        raise NotImplementedError()

    def decode(self, element):
        raise NotImplementedError()

    def __repr__(self):
        return 'GeneralizedTime({})'.format(self.name)


class TeletexString(StringType):

    def __init__(self, name):
        super(TeletexString, self).__init__(name, 'TeletexString')


class ObjectIdentifier(Type):

    def __init__(self, name):
        super(ObjectIdentifier, self).__init__(name, 'OBJECT IDENTIFIER')

    def encode(self, data):
        raise NotImplementedError()

    def decode(self, element):
        raise NotImplementedError()

    def __repr__(self):
        return 'ObjectIdentifier({})'.format(self.name)


class Choice(Type):

    def __init__(self, name, members):
        super(Choice, self).__init__(name, 'CHOICE')
        self.members = members

    def encode(self, data):
        if not isinstance(data, tuple):
            raise EncodeError("expected tuple, but got '{}'".format(data))

        for member in self.members:
            if member.name == data[0]:
                element = ElementTree.Element(self.name)
                element.append(member.encode(data[1]))

                return element

        raise EncodeError(
            "Expected choices are {}, but got '{}'.".format(
                [member.name for member in self.members],
                data[0]))

    def decode(self, element):
        for member in self.members:
            name = member.name
            member_element = element.find(name)

            if member_element is not None:
                return (name, member.decode(member_element))

        raise DecodeError('')

    def encode_of(self, data):
        if not isinstance(data, tuple):
            raise EncodeError("expected tuple, but got '{}'".format(data))

        for member in self.members:
            if member.name == data[0]:
                return member.encode(data[1])

        raise EncodeError(
            "Expected choices are {}, but got '{}'.".format(
                [member.name for member in self.members],
                data[0]))

    def decode_of(self, element):
        for member in self.members:
            if member.name == element.tag:
                return (member.name, member.decode(element))

        raise DecodeError('')

    def __repr__(self):
        return 'Choice({}, [{}])'.format(
            self.name,
            ', '.join([repr(member) for member in self.members]))


class Null(Type):

    def __init__(self, name):
        super(Null, self).__init__(name, 'NULL')

    def encode(self, data):
        return ElementTree.Element(self.name)

    def decode(self, element):
        return None

    def __repr__(self):
        return 'Null({})'.format(self.name)


class Any(Type):

    def __init__(self, name):
        super(Any, self).__init__(name, 'ANY')

    def encode(self, data):
        raise NotImplementedError()

    def decode(self, element):
        raise NotImplementedError()

    def __repr__(self):
        return 'Any({})'.format(self.name)


class Enumerated(Type):

    def __init__(self, name, values):
        super(Enumerated, self).__init__(name, 'ENUMERATED')
        self.values = enum_values_as_dict(values)

    def encode(self, data):
        for name in self.values.values():
            if data == name:
                element = ElementTree.Element(self.name)
                element.append(ElementTree.Element(data))

                return element

        raise EncodeError(
            "Enumeration value '{}' not found in {}.".format(
                data,
                [value for value in self.values.values()]))

    def decode(self, element):
        value = element[0].tag

        if value in self.values.values():
            return value

        raise DecodeError(
            "Enumeration value '{}' not found in {}.".format(
                element,
                [value for value in self.values.values()]))

    def encode_of(self, data):
        for name in self.values.values():
            if data == name:
                return ElementTree.Element(data)

        raise EncodeError(
            "Enumeration value '{}' not found in {}.".format(
                data,
                [value for value in self.values.values()]))

    def decode_of(self, element):
        value = element.tag

        if value in self.values.values():
            return value

        raise DecodeError(
            "Enumeration value '{}' not found in {}.".format(
                element,
                [value for value in self.values.values()]))

    def __repr__(self):
        return 'Enumerated({})'.format(self.name)


class Recursive(Type):

    def __init__(self, name, type_name, module_name):
        super(Recursive, self).__init__(name, 'RECURSIVE')
        self.type_name = type_name
        self.module_name = module_name

    def encode(self, data):
        raise NotImplementedError(
            "Recursive types are not yet implemented (type '{}').".format(
                self.type_name))

    def decode(self, element):
        raise NotImplementedError(
            "Recursive types are not yet implemented (type '{}').".format(
                self.type_name))

    def __repr__(self):
        return 'Recursive({})'.format(self.name)


class CompiledType(object):

    def __init__(self, type_):
        self._type = type_

    def encode(self, data):
        return ElementTree.tostring(self._type.encode(data))

    def decode(self, data):
        element = ElementTree.fromstring(data.decode('utf-8'))

        return self._type.decode(element)

    def __repr__(self):
        return repr(self._type)


class Compiler(compiler.Compiler):

    def process_type(self, type_name, type_descriptor, module_name):
        return CompiledType(self.compile_type(type_name,
                                              type_descriptor,
                                              module_name))

    def compile_type(self, name, type_descriptor, module_name):
        type_name = type_descriptor['type']

        if type_name == 'SEQUENCE':
            members = self.compile_members(
                type_descriptor['members'],
                module_name)
            compiled = Sequence(name, members)
        elif type_name == 'SEQUENCE OF':
            element = type_descriptor['element']
            compiled = SequenceOf(name,
                                  self.compile_type(element['type'],
                                                    element,
                                                    module_name))
        elif type_name == 'SET':
            members = self.compile_members(
                type_descriptor['members'],
                module_name)
            compiled = Set(name, members)
        elif type_name == 'SET OF':
            element = type_descriptor['element']
            compiled = SetOf(name,
                             self.compile_type(element['type'],
                                               element,
                                               module_name))
        elif type_name == 'CHOICE':
            members = self.compile_members(
                type_descriptor['members'],
                module_name)
            compiled = Choice(name, members)
        elif type_name == 'INTEGER':
            compiled = Integer(name)
        elif type_name == 'REAL':
            compiled = Real(name)
        elif type_name == 'ENUMERATED':
            compiled = Enumerated(name, type_descriptor['values'])
        elif type_name == 'BOOLEAN':
            compiled = Boolean(name)
        elif type_name == 'OBJECT IDENTIFIER':
            compiled = ObjectIdentifier(name)
        elif type_name == 'OCTET STRING':
            compiled = OctetString(name)
        elif type_name == 'TeletexString':
            compiled = TeletexString(name)
        elif type_name == 'NumericString':
            compiled = NumericString(name)
        elif type_name == 'PrintableString':
            compiled = PrintableString(name)
        elif type_name == 'IA5String':
            compiled = IA5String(name)
        elif type_name == 'VisibleString':
            compiled = VisibleString(name)
        elif type_name == 'GeneralString':
            compiled = GeneralString(name)
        elif type_name == 'UTF8String':
            compiled = UTF8String(name)
        elif type_name == 'BMPString':
            compiled = BMPString(name)
        elif type_name == 'GraphicString':
            compiled = GraphicString(name)
        elif type_name == 'UTCTime':
            compiled = UTCTime(name)
        elif type_name == 'UniversalString':
            compiled = UniversalString(name)
        elif type_name == 'GeneralizedTime':
            compiled = GeneralizedTime(name)
        elif type_name == 'BIT STRING':
            compiled = BitString(name)
        elif type_name == 'ANY':
            compiled = Any(name)
        elif type_name == 'ANY DEFINED BY':
            compiled = Any(name)
        elif type_name == 'NULL':
            compiled = Null(name)
        else:
            if type_name in self.types_backtrace:
                compiled = Recursive(name,
                                     type_name,
                                     module_name)
            else:
                self.types_backtrace_push(type_name)
                compiled = self.compile_type(
                    name,
                    *self.lookup_type_descriptor(
                        type_name,
                        module_name))
                self.types_backtrace_pop()

        return compiled

    def compile_members(self, members, module_name):
        compiled_members = []

        for member in members:
            if member == '...':
                continue

            if isinstance(member, list):
                compiled_members.extend(self.compile_members(member,
                                                             module_name))
                continue

            compiled_member = self.compile_type(member['name'],
                                                member,
                                                module_name)

            if 'optional' in member:
                compiled_member.optional = member['optional']

            if 'default' in member:
                compiled_member.default = member['default']

            compiled_members.append(compiled_member)

        return compiled_members


def compile_dict(specification):
    return Compiler(specification).process()


def decode_length(_data):
    raise DecodeError('Decode length is not supported for this codec.')



EXPECTED = {'X691-A3': {'extensibility-implied': False,
             'imports': {},
             'object-classes': {},
             'object-sets': {},
             'types': {'ChildInformation': {'members': [{'name': 'name',
                                                         'type': 'Name'},
                                                        {'name': 'dateOfBirth',
                                                         'tag': {'number': 0},
                                                         'type': 'Date'},
                                                        '...',
                                                        {'name': 'sex',
                                                         'optional': True,
                                                         'tag': {'kind': 'IMPLICIT',
                                                                 'number': 1},
                                                         'type': 'ENUMERATED',
                                                         'values': [('male', 1),
                                                                    ('female',
                                                                     2),
                                                                    ('unknown',
                                                                     3)]}],
                                            'type': 'SET'},
                       'Date': {'tag': {'class': 'APPLICATION',
                                        'kind': 'IMPLICIT',
                                        'number': 3},
                                'type': 'VisibleString'},
                       'EmployeeNumber': {'restricted-to': [(0, 9999), '...'],
                                          'tag': {'class': 'APPLICATION',
                                                  'kind': 'IMPLICIT',
                                                  'number': 2},
                                          'type': 'INTEGER'},
                       'Name': {'members': [{'name': 'givenName',
                                             'type': 'NameString'},
                                            {'name': 'initial',
                                             'type': 'NameString'},
                                            {'name': 'familyName',
                                             'type': 'NameString'},
                                            '...'],
                                'tag': {'class': 'APPLICATION',
                                        'kind': 'IMPLICIT',
                                        'number': 1},
                                'type': 'SEQUENCE'},
                       'NameString': {'type': 'VisibleString'},
                       'PersonnelRecord': {'members': [{'name': 'name',
                                                        'type': 'Name'},
                                                       {'name': 'title',
                                                        'tag': {'number': 0},
                                                        'type': 'VisibleString'},
                                                       {'name': 'number',
                                                        'type': 'EmployeeNumber'},
                                                       {'name': 'dateOfHire',
                                                        'tag': {'number': 1},
                                                        'type': 'Date'},
                                                       {'name': 'nameOfSpouse',
                                                        'tag': {'number': 2},
                                                        'type': 'Name'},
                                                       {'element': {'type': 'ChildInformation'},
                                                        'name': 'children',
                                                        'optional': True,
                                                        'size': [2, '...'],
                                                        'tag': {'kind': 'IMPLICIT',
                                                                'number': 3},
                                                        'type': 'SEQUENCE OF'},
                                                       '...'],
                                           'tag': {'class': 'APPLICATION',
                                                   'kind': 'IMPLICIT',
                                                   'number': 0},
                                           'type': 'SET'}},
             'values': {}}}


EXPECTED = {'PKIXCMP': {'extensibility-implied': False,
             'imports': {'PKCS-10': ['CertificationRequest'],
                         'PKIX1Explicit88': ['Certificate',
                                             'CertificateList',
                                             'Extensions',
                                             'AlgorithmIdentifier',
                                             'UTF8String'],
                         'PKIX1Implicit88': ['GeneralName', 'KeyIdentifier'],
                         'PKIXCRMF-2005': ['CertTemplate',
                                           'PKIPublicationInfo',
                                           'EncryptedValue',
                                           'CertId',
                                           'CertReqMessages']},
             'object-classes': {},
             'object-sets': {},
             'tags': 'EXPLICIT',
             'types': {'CAKeyUpdAnnContent': {'members': [{'name': 'oldWithNew',
                                                           'type': 'CMPCertificate'},
                                                          {'name': 'newWithOld',
                                                           'type': 'CMPCertificate'},
                                                          {'name': 'newWithNew',
                                                           'type': 'CMPCertificate'}],
                                              'type': 'SEQUENCE'},
                       'CMPCertificate': {'members': [{'name': 'x509v3PKCert',
                                                       'type': 'Certificate'}],
                                          'type': 'CHOICE'},
                       'CRLAnnContent': {'element': {'type': 'CertificateList'},
                                         'type': 'SEQUENCE OF'},
                       'CertAnnContent': {'type': 'CMPCertificate'},
                       'CertConfirmContent': {'element': {'type': 'CertStatus'},
                                              'type': 'SEQUENCE OF'},
                       'CertOrEncCert': {'members': [{'name': 'certificate',
                                                      'tag': {'number': 0},
                                                      'type': 'CMPCertificate'},
                                                     {'name': 'encryptedCert',
                                                      'tag': {'number': 1},
                                                      'type': 'EncryptedValue'}],
                                         'type': 'CHOICE'},
                       'CertRepMessage': {'members': [{'element': {'type': 'CMPCertificate'},
                                                       'name': 'caPubs',
                                                       'optional': True,
                                                       'size': [(1, 'MAX')],
                                                       'tag': {'number': 1},
                                                       'type': 'SEQUENCE OF'},
                                                      {'element': {'type': 'CertResponse'},
                                                       'name': 'response',
                                                       'type': 'SEQUENCE OF'}],
                                          'type': 'SEQUENCE'},
                       'CertResponse': {'members': [{'name': 'certReqId',
                                                     'type': 'INTEGER'},
                                                    {'name': 'status',
                                                     'type': 'PKIStatusInfo'},
                                                    {'name': 'certifiedKeyPair',
                                                     'optional': True,
                                                     'type': 'CertifiedKeyPair'},
                                                    {'name': 'rspInfo',
                                                     'optional': True,
                                                     'type': 'OCTET STRING'}],
                                        'type': 'SEQUENCE'},
                       'CertStatus': {'members': [{'name': 'certHash',
                                                   'type': 'OCTET STRING'},
                                                  {'name': 'certReqId',
                                                   'type': 'INTEGER'},
                                                  {'name': 'statusInfo',
                                                   'optional': True,
                                                   'type': 'PKIStatusInfo'}],
                                      'type': 'SEQUENCE'},
                       'CertifiedKeyPair': {'members': [{'name': 'certOrEncCert',
                                                         'type': 'CertOrEncCert'},
                                                        {'name': 'privateKey',
                                                         'optional': True,
                                                         'tag': {'number': 0},
                                                         'type': 'EncryptedValue'},
                                                        {'name': 'publicationInfo',
                                                         'optional': True,
                                                         'tag': {'number': 1},
                                                         'type': 'PKIPublicationInfo'}],
                                            'type': 'SEQUENCE'},
                       'Challenge': {'members': [{'name': 'owf',
                                                  'optional': True,
                                                  'type': 'AlgorithmIdentifier'},
                                                 {'name': 'witness',
                                                  'type': 'OCTET STRING'},
                                                 {'name': 'challenge',
                                                  'type': 'OCTET STRING'}],
                                     'type': 'SEQUENCE'},
                       'DHBMParameter': {'members': [{'name': 'owf',
                                                      'type': 'AlgorithmIdentifier'},
                                                     {'name': 'mac',
                                                      'type': 'AlgorithmIdentifier'}],
                                         'type': 'SEQUENCE'},
                       'ErrorMsgContent': {'members': [{'name': 'pKIStatusInfo',
                                                        'type': 'PKIStatusInfo'},
                                                       {'name': 'errorCode',
                                                        'optional': True,
                                                        'type': 'INTEGER'},
                                                       {'name': 'errorDetails',
                                                        'optional': True,
                                                        'type': 'PKIFreeText'}],
                                           'type': 'SEQUENCE'},
                       'GenMsgContent': {'element': {'type': 'InfoTypeAndValue'},
                                         'type': 'SEQUENCE OF'},
                       'GenRepContent': {'element': {'type': 'InfoTypeAndValue'},
                                         'type': 'SEQUENCE OF'},
                       'InfoTypeAndValue': {'members': [{'name': 'infoType',
                                                         'type': 'OBJECT '
                                                                 'IDENTIFIER'},
                                                        {'choices': {},
                                                         'name': 'infoValue',
                                                         'optional': True,
                                                         'type': 'ANY DEFINED '
                                                                 'BY',
                                                         'value': 'infoType'}],
                                            'type': 'SEQUENCE'},
                       'KeyRecRepContent': {'members': [{'name': 'status',
                                                         'type': 'PKIStatusInfo'},
                                                        {'name': 'newSigCert',
                                                         'optional': True,
                                                         'tag': {'number': 0},
                                                         'type': 'CMPCertificate'},
                                                        {'element': {'type': 'CMPCertificate'},
                                                         'name': 'caCerts',
                                                         'optional': True,
                                                         'size': [(1, 'MAX')],
                                                         'tag': {'number': 1},
                                                         'type': 'SEQUENCE OF'},
                                                        {'element': {'type': 'CertifiedKeyPair'},
                                                         'name': 'keyPairHist',
                                                         'optional': True,
                                                         'size': [(1, 'MAX')],
                                                         'tag': {'number': 2},
                                                         'type': 'SEQUENCE '
                                                                 'OF'}],
                                            'type': 'SEQUENCE'},
                       'NestedMessageContent': {'type': 'PKIMessages'},
                       'OOBCert': {'type': 'CMPCertificate'},
                       'OOBCertHash': {'members': [{'name': 'hashAlg',
                                                    'optional': True,
                                                    'tag': {'number': 0},
                                                    'type': 'AlgorithmIdentifier'},
                                                   {'name': 'certId',
                                                    'optional': True,
                                                    'tag': {'number': 1},
                                                    'type': 'CertId'},
                                                   {'name': 'hashVal',
                                                    'type': 'BIT STRING'}],
                                       'type': 'SEQUENCE'},
                       'PBMParameter': {'members': [{'name': 'salt',
                                                     'type': 'OCTET STRING'},
                                                    {'name': 'owf',
                                                     'type': 'AlgorithmIdentifier'},
                                                    {'name': 'iterationCount',
                                                     'type': 'INTEGER'},
                                                    {'name': 'mac',
                                                     'type': 'AlgorithmIdentifier'}],
                                        'type': 'SEQUENCE'},
                       'PKIBody': {'members': [{'name': 'ir',
                                                'tag': {'number': 0},
                                                'type': 'CertReqMessages'},
                                               {'name': 'ip',
                                                'tag': {'number': 1},
                                                'type': 'CertRepMessage'},
                                               {'name': 'cr',
                                                'tag': {'number': 2},
                                                'type': 'CertReqMessages'},
                                               {'name': 'cp',
                                                'tag': {'number': 3},
                                                'type': 'CertRepMessage'},
                                               {'name': 'p10cr',
                                                'tag': {'number': 4},
                                                'type': 'CertificationRequest'},
                                               {'name': 'popdecc',
                                                'tag': {'number': 5},
                                                'type': 'POPODecKeyChallContent'},
                                               {'name': 'popdecr',
                                                'tag': {'number': 6},
                                                'type': 'POPODecKeyRespContent'},
                                               {'name': 'kur',
                                                'tag': {'number': 7},
                                                'type': 'CertReqMessages'},
                                               {'name': 'kup',
                                                'tag': {'number': 8},
                                                'type': 'CertRepMessage'},
                                               {'name': 'krr',
                                                'tag': {'number': 9},
                                                'type': 'CertReqMessages'},
                                               {'name': 'krp',
                                                'tag': {'number': 10},
                                                'type': 'KeyRecRepContent'},
                                               {'name': 'rr',
                                                'tag': {'number': 11},
                                                'type': 'RevReqContent'},
                                               {'name': 'rp',
                                                'tag': {'number': 12},
                                                'type': 'RevRepContent'},
                                               {'name': 'ccr',
                                                'tag': {'number': 13},
                                                'type': 'CertReqMessages'},
                                               {'name': 'ccp',
                                                'tag': {'number': 14},
                                                'type': 'CertRepMessage'},
                                               {'name': 'ckuann',
                                                'tag': {'number': 15},
                                                'type': 'CAKeyUpdAnnContent'},
                                               {'name': 'cann',
                                                'tag': {'number': 16},
                                                'type': 'CertAnnContent'},
                                               {'name': 'rann',
                                                'tag': {'number': 17},
                                                'type': 'RevAnnContent'},
                                               {'name': 'crlann',
                                                'tag': {'number': 18},
                                                'type': 'CRLAnnContent'},
                                               {'name': 'pkiconf',
                                                'tag': {'number': 19},
                                                'type': 'PKIConfirmContent'},
                                               {'name': 'nested',
                                                'tag': {'number': 20},
                                                'type': 'NestedMessageContent'},
                                               {'name': 'genm',
                                                'tag': {'number': 21},
                                                'type': 'GenMsgContent'},
                                               {'name': 'genp',
                                                'tag': {'number': 22},
                                                'type': 'GenRepContent'},
                                               {'name': 'error',
                                                'tag': {'number': 23},
                                                'type': 'ErrorMsgContent'},
                                               {'name': 'certConf',
                                                'tag': {'number': 24},
                                                'type': 'CertConfirmContent'},
                                               {'name': 'pollReq',
                                                'tag': {'number': 25},
                                                'type': 'PollReqContent'},
                                               {'name': 'pollRep',
                                                'tag': {'number': 26},
                                                'type': 'PollRepContent'}],
                                   'type': 'CHOICE'},
                       'PKIConfirmContent': {'type': 'NULL'},
                       'PKIFailureInfo': {'type': 'BIT STRING'},
                       'PKIFreeText': {'element': {'type': 'UTF8String'},
                                       'size': [(1, 'MAX')],
                                       'type': 'SEQUENCE OF'},
                       'PKIHeader': {'members': [{'name': 'pvno',
                                                  'type': 'INTEGER'},
                                                 {'name': 'sender',
                                                  'type': 'GeneralName'},
                                                 {'name': 'recipient',
                                                  'type': 'GeneralName'},
                                                 {'name': 'messageTime',
                                                  'optional': True,
                                                  'tag': {'number': 0},
                                                  'type': 'GeneralizedTime'},
                                                 {'name': 'protectionAlg',
                                                  'optional': True,
                                                  'tag': {'number': 1},
                                                  'type': 'AlgorithmIdentifier'},
                                                 {'name': 'senderKID',
                                                  'optional': True,
                                                  'tag': {'number': 2},
                                                  'type': 'KeyIdentifier'},
                                                 {'name': 'recipKID',
                                                  'optional': True,
                                                  'tag': {'number': 3},
                                                  'type': 'KeyIdentifier'},
                                                 {'name': 'transactionID',
                                                  'optional': True,
                                                  'tag': {'number': 4},
                                                  'type': 'OCTET STRING'},
                                                 {'name': 'senderNonce',
                                                  'optional': True,
                                                  'tag': {'number': 5},
                                                  'type': 'OCTET STRING'},
                                                 {'name': 'recipNonce',
                                                  'optional': True,
                                                  'tag': {'number': 6},
                                                  'type': 'OCTET STRING'},
                                                 {'name': 'freeText',
                                                  'optional': True,
                                                  'tag': {'number': 7},
                                                  'type': 'PKIFreeText'},
                                                 {'element': {'type': 'InfoTypeAndValue'},
                                                  'name': 'generalInfo',
                                                  'optional': True,
                                                  'size': [(1, 'MAX')],
                                                  'tag': {'number': 8},
                                                  'type': 'SEQUENCE OF'}],
                                     'type': 'SEQUENCE'},
                       'PKIMessage': {'members': [{'name': 'header',
                                                   'type': 'PKIHeader'},
                                                  {'name': 'body',
                                                   'type': 'PKIBody'},
                                                  {'name': 'protection',
                                                   'optional': True,
                                                   'tag': {'number': 0},
                                                   'type': 'PKIProtection'},
                                                  {'element': {'type': 'CMPCertificate'},
                                                   'name': 'extraCerts',
                                                   'optional': True,
                                                   'size': [(1, 'MAX')],
                                                   'tag': {'number': 1},
                                                   'type': 'SEQUENCE OF'}],
                                      'type': 'SEQUENCE'},
                       'PKIMessages': {'element': {'type': 'PKIMessage'},
                                       'size': [(1, 'MAX')],
                                       'type': 'SEQUENCE OF'},
                       'PKIProtection': {'type': 'BIT STRING'},
                       'PKIStatus': {'type': 'INTEGER'},
                       'PKIStatusInfo': {'members': [{'name': 'status',
                                                      'type': 'PKIStatus'},
                                                     {'name': 'statusString',
                                                      'optional': True,
                                                      'type': 'PKIFreeText'},
                                                     {'name': 'failInfo',
                                                      'optional': True,
                                                      'type': 'PKIFailureInfo'}],
                                         'type': 'SEQUENCE'},
                       'POPODecKeyChallContent': {'element': {'type': 'Challenge'},
                                                  'type': 'SEQUENCE OF'},
                       'POPODecKeyRespContent': {'element': {'type': 'INTEGER'},
                                                 'type': 'SEQUENCE OF'},
                       'PollRepContent': {'element': {'members': [{'name': 'certReqId',
                                                                   'type': 'INTEGER'},
                                                                  {'name': 'checkAfter',
                                                                   'type': 'INTEGER'},
                                                                  {'name': 'reason',
                                                                   'optional': True,
                                                                   'type': 'PKIFreeText'}],
                                                      'type': 'SEQUENCE'},
                                          'type': 'SEQUENCE OF'},
                       'PollReqContent': {'element': {'members': [{'name': 'certReqId',
                                                                   'type': 'INTEGER'}],
                                                      'type': 'SEQUENCE'},
                                          'type': 'SEQUENCE OF'},
                       'ProtectedPart': {'members': [{'name': 'header',
                                                      'type': 'PKIHeader'},
                                                     {'name': 'body',
                                                      'type': 'PKIBody'}],
                                         'type': 'SEQUENCE'},
                       'RevAnnContent': {'members': [{'name': 'status',
                                                      'type': 'PKIStatus'},
                                                     {'name': 'certId',
                                                      'type': 'CertId'},
                                                     {'name': 'willBeRevokedAt',
                                                      'type': 'GeneralizedTime'},
                                                     {'name': 'badSinceDate',
                                                      'type': 'GeneralizedTime'},
                                                     {'name': 'crlDetails',
                                                      'optional': True,
                                                      'type': 'Extensions'}],
                                         'type': 'SEQUENCE'},
                       'RevDetails': {'members': [{'name': 'certDetails',
                                                   'type': 'CertTemplate'},
                                                  {'name': 'crlEntryDetails',
                                                   'optional': True,
                                                   'type': 'Extensions'}],
                                      'type': 'SEQUENCE'},
                       'RevRepContent': {'members': [{'element': {'type': 'PKIStatusInfo'},
                                                      'name': 'status',
                                                      'size': [(1, 'MAX')],
                                                      'type': 'SEQUENCE OF'},
                                                     {'element': {'type': 'CertId'},
                                                      'name': 'revCerts',
                                                      'optional': True,
                                                      'size': [(1, 'MAX')],
                                                      'tag': {'number': 0},
                                                      'type': 'SEQUENCE OF'},
                                                     {'element': {'type': 'CertificateList'},
                                                      'name': 'crls',
                                                      'optional': True,
                                                      'size': [(1, 'MAX')],
                                                      'tag': {'number': 1},
                                                      'type': 'SEQUENCE OF'}],
                                         'type': 'SEQUENCE'},
                       'RevReqContent': {'element': {'type': 'RevDetails'},
                                         'type': 'SEQUENCE OF'}},
             'values': {'id-DHBasedMac': {'type': 'OBJECT IDENTIFIER',
                                          'value': [1,
                                                    2,
                                                    840,
                                                    113533,
                                                    7,
                                                    66,
                                                    30]},
                        'id-PasswordBasedMac': {'type': 'OBJECT IDENTIFIER',
                                                'value': [1,
                                                          2,
                                                          840,
                                                          113533,
                                                          7,
                                                          66,
                                                          13]}}}}


EXPECTED = {'PKIXTSP': {'extensibility-implied': False,
             'imports': {'CryptographicMessageSyntax': ['ContentInfo'],
                         'PKIX1Explicit88': ['Extensions',
                                             'AlgorithmIdentifier'],
                         'PKIX1Implicit88': ['GeneralName'],
                         'PKIXCMP': ['PKIFreeText']},
             'object-classes': {},
             'object-sets': {},
             'tags': 'IMPLICIT',
             'types': {'Accuracy': {'members': [{'name': 'seconds',
                                                 'optional': True,
                                                 'type': 'INTEGER'},
                                                {'name': 'millis',
                                                 'optional': True,
                                                 'restricted-to': [(1, 999)],
                                                 'tag': {'number': 0},
                                                 'type': 'INTEGER'},
                                                {'name': 'micros',
                                                 'optional': True,
                                                 'restricted-to': [(1, 999)],
                                                 'tag': {'number': 1},
                                                 'type': 'INTEGER'}],
                                    'type': 'SEQUENCE'},
                       'MessageImprint': {'members': [{'name': 'hashAlgorithm',
                                                       'type': 'AlgorithmIdentifier'},
                                                      {'name': 'hashedMessage',
                                                       'type': 'OCTET STRING'}],
                                          'type': 'SEQUENCE'},
                       'PKIFailureInfo': {'type': 'BIT STRING'},
                       'PKIStatus': {'type': 'INTEGER'},
                       'PKIStatusInfo': {'members': [{'name': 'status',
                                                      'type': 'PKIStatus'},
                                                     {'name': 'statusString',
                                                      'optional': True,
                                                      'type': 'PKIFreeText'},
                                                     {'name': 'failInfo',
                                                      'optional': True,
                                                      'type': 'PKIFailureInfo'}],
                                         'type': 'SEQUENCE'},
                       'TSAPolicyId': {'type': 'OBJECT IDENTIFIER'},
                       'TSTInfo': {'members': [{'name': 'version',
                                                'type': 'INTEGER'},
                                               {'name': 'policy',
                                                'type': 'TSAPolicyId'},
                                               {'name': 'messageImprint',
                                                'type': 'MessageImprint'},
                                               {'name': 'serialNumber',
                                                'type': 'INTEGER'},
                                               {'name': 'genTime',
                                                'type': 'GeneralizedTime'},
                                               {'name': 'accuracy',
                                                'optional': True,
                                                'type': 'Accuracy'},
                                               {'default': False,
                                                'name': 'ordering',
                                                'type': 'BOOLEAN'},
                                               {'name': 'nonce',
                                                'optional': True,
                                                'type': 'INTEGER'},
                                               {'name': 'tsa',
                                                'optional': True,
                                                'tag': {'number': 0},
                                                'type': 'GeneralName'},
                                               {'name': 'extensions',
                                                'optional': True,
                                                'tag': {'kind': 'IMPLICIT',
                                                        'number': 1},
                                                'type': 'Extensions'}],
                                   'type': 'SEQUENCE'},
                       'TimeStampReq': {'members': [{'name': 'version',
                                                     'type': 'INTEGER'},
                                                    {'name': 'messageImprint',
                                                     'type': 'MessageImprint'},
                                                    {'name': 'reqPolicy',
                                                     'optional': True,
                                                     'type': 'TSAPolicyId'},
                                                    {'name': 'nonce',
                                                     'optional': True,
                                                     'type': 'INTEGER'},
                                                    {'default': False,
                                                     'name': 'certReq',
                                                     'type': 'BOOLEAN'},
                                                    {'name': 'extensions',
                                                     'optional': True,
                                                     'tag': {'kind': 'IMPLICIT',
                                                             'number': 0},
                                                     'type': 'Extensions'}],
                                        'type': 'SEQUENCE'},
                       'TimeStampResp': {'members': [{'name': 'status',
                                                      'type': 'PKIStatusInfo'},
                                                     {'name': 'timeStampToken',
                                                      'optional': True,
                                                      'type': 'TimeStampToken'}],
                                         'type': 'SEQUENCE'},
                       'TimeStampToken': {'type': 'ContentInfo'}},
             'values': {'id-ct-TSTInfo': {'type': 'OBJECT IDENTIFIER',
                                          'value': [('iso', 1),
                                                    ('member-body', 2),
                                                    ('us', 840),
                                                    ('rsadsi', 113549),
                                                    ('pkcs', 1),
                                                    ('pkcs-9', 9),
                                                    ('smime', 16),
                                                    ('ct', 1),
                                                    4]}}}}


import unittest
from .utils import Asn1ToolsBaseTest
import asn1tools
import sys
from copy import deepcopy

sys.path.append('tests/files')
sys.path.append('tests/files/3gpp')

from rrc_8_6_0 import EXPECTED as RRC_8_6_0
from s1ap_14_4_0 import EXPECTED as S1AP_14_4_0


class Asn1ToolsPerTest(Asn1ToolsBaseTest):

    maxDiff = None

    def test_foo(self):
        foo = asn1tools.compile_files('tests/files/foo.asn', 'per')

        self.assertEqual(len(foo.types), 2)
        self.assertTrue(foo.types['Question'] is not None)
        self.assertTrue(foo.types['Answer'] is not None)
        self.assertEqual(len(foo.modules), 1)
        self.assertTrue(foo.modules['Foo'] is not None)

        # Encode a question.
        encoded = foo.encode('Question',
                             {'id': 1, 'question': 'Is 1+1=3?'})
        self.assertEqual(encoded,
                         b'\x01\x01\x09\x49\x73\x20\x31\x2b\x31\x3d\x33\x3f')

        # Decode the encoded question.
        decoded = foo.decode('Question', encoded)
        self.assertEqual(decoded, {'id': 1, 'question': 'Is 1+1=3?'})

        # Encode an answer.
        encoded = foo.encode('Answer', {'id': 1, 'answer': False})
        self.assertEqual(encoded, b'\x01\x01\x00')

        # Decode the encoded answer.
        decoded = foo.decode('Answer', encoded)
        self.assertEqual(decoded, {'id': 1, 'answer': False})

        # Encode a question with missing field 'id'.
        with self.assertRaises(asn1tools.EncodeError) as cm:
            encoded = foo.encode('Question', {'question': 'Is 1+1=3?'})

        self.assertEqual(
            str(cm.exception),
            "Sequence member 'id' not found in {'question': 'Is 1+1=3?'}.")

    def test_decode_length(self):
        foo = asn1tools.compile_files('tests/files/foo.asn', 'per')

        with self.assertRaises(asn1tools.DecodeError) as cm:
            foo.decode_length(b'')

        self.assertEqual(str(cm.exception),
                         ': Decode length is not supported for this codec.')

    def test_x691_a1(self):
        a1 = asn1tools.compile_files('tests/files/x691_a1.asn', 'per')

        decoded = {
            'name': {
                'givenName': 'John',
                'initial': 'P',
                'familyName': 'Smith'
            },
            'title': 'Director',
            'number': 51,
            'dateOfHire': '19710917',
            'nameOfSpouse': {
                'givenName': 'Mary',
                'initial': 'T',
                'familyName': 'Smith'
            },
            'children': [
                {
                    'name': {
                        'givenName': 'Ralph',
                        'initial': 'T',
                        'familyName': 'Smith'
                    },
                    'dateOfBirth': '19571111'
                },
                {
                    'name': {
                        'givenName': 'Susan',
                        'initial': 'B',
                        'familyName': 'Jones'
                    },
                    'dateOfBirth': '19590717'
                }
            ]
        }

        encoded = (
            b'\x80\x04\x4a\x6f\x68\x6e\x01\x50\x05\x53\x6d\x69\x74\x68\x01\x33'
            b'\x08\x44\x69\x72\x65\x63\x74\x6f\x72\x08\x31\x39\x37\x31\x30\x39'
            b'\x31\x37\x04\x4d\x61\x72\x79\x01\x54\x05\x53\x6d\x69\x74\x68\x02'
            b'\x05\x52\x61\x6c\x70\x68\x01\x54\x05\x53\x6d\x69\x74\x68\x08\x31'
            b'\x39\x35\x37\x31\x31\x31\x31\x05\x53\x75\x73\x61\x6e\x01\x42\x05'
            b'\x4a\x6f\x6e\x65\x73\x08\x31\x39\x35\x39\x30\x37\x31\x37'
        )

        with self.assertRaises(NotImplementedError):
            self.assert_encode_decode(a1, 'PersonnelRecord', decoded, encoded)

    def test_x691_a2(self):
        a2 = asn1tools.compile_files('tests/files/x691_a2.asn', 'per')

        decoded = {
            'name': {
                'givenName': 'John',
                'initial': 'P',
                'familyName': 'Smith'
            },
            'title': 'Director',
            'number': 51,
            'dateOfHire': '19710917',
            'nameOfSpouse': {
                'givenName': 'Mary',
                'initial': 'T',
                'familyName': 'Smith'
            },
            'children': [
                {
                    'name': {
                        'givenName': 'Ralph',
                        'initial': 'T',
                        'familyName': 'Smith'
                    },
                    'dateOfBirth': '19571111'
                },
                {
                    'name': {
                        'givenName': 'Susan',
                        'initial': 'B',
                        'familyName': 'Jones'
                    },
                    'dateOfBirth': '19590717'
                }
            ]
        }

        encoded = (
            b'\x86\x4a\x6f\x68\x6e\x50\x10\x53\x6d\x69\x74\x68\x01\x33\x08\x44'
            b'\x69\x72\x65\x63\x74\x6f\x72\x19\x71\x09\x17\x0c\x4d\x61\x72\x79'
            b'\x54\x10\x53\x6d\x69\x74\x68\x02\x10\x52\x61\x6c\x70\x68\x54\x10'
            b'\x53\x6d\x69\x74\x68\x19\x57\x11\x11\x10\x53\x75\x73\x61\x6e\x42'
            b'\x10\x4a\x6f\x6e\x65\x73\x19\x59\x07\x17'
        )

        with self.assertRaises(NotImplementedError):
            self.assert_encode_decode(a2, 'PersonnelRecord', decoded, encoded)

    def test_x691_a3(self):
        a3 = asn1tools.compile_files('tests/files/x691_a3.asn', 'per')
        
        return

        decoded = {
            'name': {
                'givenName': 'John',
                'initial': 'P',
                'familyName': 'Smith'
            },
            'title': 'Director',
            'number': 51,
            'dateOfHire': '19710917',
            'nameOfSpouse': {
                'givenName': 'Mary',
                'initial': 'T',
                'familyName': 'Smith'
            },
            'children': [
                {
                    'name': {
                        'givenName': 'Ralph',
                        'initial': 'T',
                        'familyName': 'Smith'
                    },
                    'dateOfBirth': '19571111'
                },
                {
                    'name': {
                        'givenName': 'Susan',
                        'initial': 'B',
                        'familyName': 'Jones'
                    },
                    'dateOfBirth': '19590717',
                    'sex': 'female'
                }
            ]
        }

        encoded = (
            b'\x40\xc0\x4a\x6f\x68\x6e\x50\x08\x53\x6d\x69\x74\x68\x00\x00\x33'
            b'\x08\x44\x69\x72\x65\x63\x74\x6f\x72\x00\x19\x71\x09\x17\x03\x4d'
            b'\x61\x72\x79\x54\x08\x53\x6d\x69\x74\x68\x01\x00\x52\x61\x6c\x70'
            b'\x68\x54\x08\x53\x6d\x69\x74\x68\x00\x19\x57\x11\x11\x82\x00\x53'
            b'\x75\x73\x61\x6e\x42\x08\x4a\x6f\x6e\x65\x73\x00\x19\x59\x07\x17'
            b'\x01\x01\x40'
        )

        self.assert_encode_decode(a3, 'PersonnelRecord', decoded, encoded)

    def test_x691_a4(self):
        a4 = asn1tools.compile_files('tests/files/x691_a4.asn', 'per')

        decoded = {
            'a': 253,
            'b': True,
            'c': (
                'e', True
            ),
            'g': '123',
            'h': True
        }

        encoded = (
            b'\x9e\x00\x01\x80\x01\x02\x91\xa4'
        )

        with self.assertRaises(NotImplementedError):
            self.assert_encode_decode(a4, 'Ax', decoded, encoded)

    def test_rrc_8_6_0(self):
        rrc = asn1tools.compile_dict(deepcopy(RRC_8_6_0), 'per')

        # Message 1.
        encoded = rrc.encode('PCCH-Message',
                             {
                                 'message': (
                                     'c1',
                                     (
                                         'paging',
                                         {
                                             'systemInfoModification': 'true',
                                             'nonCriticalExtension': {
                                             }
                                         }
                                     )
                                 )
                             })
        self.assertEqual(encoded, b'\x28')

        # Message 2.
        encoded = rrc.encode('PCCH-Message',
                             {
                                 'message': (
                                     'c1',
                                     (
                                         'paging', {
                                         }
                                     )
                                 )
                             })
        self.assertEqual(encoded, b'\x00')

        # Message 3.
        encoded = rrc.encode('BCCH-BCH-Message',
                             {
                                 'message': {
                                     'dl-Bandwidth': 'n6',
                                     'phich-Config': {
                                         'phich-Duration': 'normal',
                                         'phich-Resource': 'half'
                                     },
                                     'systemFrameNumber': (b'\x12', 8),
                                     'spare': (b'\x34\x56', 10)
                                 }
                             })
        self.assertEqual(encoded, b'\x04\x48\xd1')

    def test_encode_all_types(self):
        all_types = asn1tools.compile_files('tests/files/all_types.asn',
                                            'per')

        self.assertEqual(all_types.encode('Bitstring', (b'\x40', 4)),
                         b'\x04\x40')
        self.assertEqual(all_types.encode('Octetstring', b'\x00'),
                         b'\x01\x00')
        self.assertEqual(all_types.encode('Null', None), b'')
        self.assertEqual(all_types.encode('Objectidentifier', '1.2'),
                         b'\x01\x2a')
        self.assertEqual(all_types.encode('Objectidentifier', '1.2.333'),
                         b'\x03\x2a\x82\x4d')
        self.assertEqual(all_types.encode('Enumerated', 'one'), b'\x00')

        with self.assertRaises(NotImplementedError):
            all_types.encode('Utf8string', 'foo')

        self.assertEqual(all_types.encode('Sequence', {}), b'')

        with self.assertRaises(NotImplementedError):
            all_types.encode('Sequence12', {'a': [{'a': []}]})

        with self.assertRaises(NotImplementedError):
            all_types.encode('Set', {})

        with self.assertRaises(NotImplementedError):
            all_types.encode('Numericstring', '123')

        with self.assertRaises(NotImplementedError):
            all_types.encode('Printablestring', 'foo')

        self.assertEqual(all_types.encode('Ia5string', 'bar'), b'\x03bar')

        with self.assertRaises(NotImplementedError):
            all_types.encode('Universalstring', 'bar')

        with self.assertRaises(NotImplementedError):
            all_types.encode('Visiblestring', 'bar')

        with self.assertRaises(NotImplementedError):
            all_types.encode('Generalstring', 'bar')

        with self.assertRaises(NotImplementedError):
            all_types.encode('Bmpstring', b'bar')

        with self.assertRaises(NotImplementedError):
            all_types.encode('Teletexstring', b'fum')

        with self.assertRaises(NotImplementedError):
            all_types.encode('Utctime', '010203040506')

    def test_decode_all_types(self):
        all_types = asn1tools.compile_files('tests/files/all_types.asn',
                                            'per')

        self.assertEqual(all_types.decode('Bitstring', b'\x04\x40'),
                         (b'\x40', 4))
        self.assertEqual(all_types.decode('Octetstring', b'\x01\x00'),
                         b'\x00')
        self.assertEqual(all_types.decode('Null', b'\x05\x00'), None)
        self.assertEqual(all_types.decode('Objectidentifier', b'\x01\x2a'),
                         '1.2')
        self.assertEqual(all_types.decode('Objectidentifier', b'\x03\x2a\x82\x4d'),
                         '1.2.333')
        self.assertEqual(all_types.decode('Enumerated', b'\x00'), 'one')

        with self.assertRaises(NotImplementedError):
            all_types.decode('Utf8string', b'\x0c\x03foo')

        self.assertEqual(all_types.decode('Sequence', b''), {})

        with self.assertRaises(NotImplementedError):
            all_types.decode('Sequence12', b'\x80\x01\x00')

        with self.assertRaises(NotImplementedError):
            all_types.decode('Set', b'\x31\x00')

        with self.assertRaises(NotImplementedError):
            all_types.decode('Numericstring', b'\x12\x03123')

        with self.assertRaises(NotImplementedError):
            all_types.decode('Printablestring', b'\x13\x03foo')

        self.assertEqual(all_types.decode('Ia5string', b'\x03bar'), 'bar')

        with self.assertRaises(NotImplementedError):
            all_types.decode('Universalstring', b'\x1c\x03bar')

        with self.assertRaises(NotImplementedError):
            all_types.decode('Visiblestring', b'\x1a\x03bar')

        with self.assertRaises(NotImplementedError):
            all_types.decode('Generalstring', b'\x1b\x03bar')

        with self.assertRaises(NotImplementedError):
            all_types.decode('Bmpstring', b'\x1e\x03bar')

        with self.assertRaises(NotImplementedError):
            all_types.decode('Teletexstring', b'\x14\x03fum')

        with self.assertRaises(NotImplementedError):
            all_types.decode('Utctime', b'\x17\x0d010203040506Y')

    def test_all_types_automatic_tags(self):
        all_types = asn1tools.compile_files(
            'tests/files/all_types_automatic_tags.asn', 'per')

        datas = [
            ('Sequence3', {'a': 1, 'c': 2,'d': True}, b'\x00\x01\x01\x01\x02\x80')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(all_types, type_name, decoded, encoded)

    def test_bar(self):
        """A simple example.

        """

        bar = asn1tools.compile_files('tests/files/bar.asn', 'per')

        # Message 1.
        decoded = {
            'headerOnly': True,
            'lock': False,
            'acceptTypes': {
                'standardTypes': [(b'\x40', 2), (b'\x80', 1)]
            },
            'url': b'/ses/magic/moxen.html'
        }

        encoded = (
            b'\xd0\x02\x02\x40\x01\x80\x15\x2f\x73\x65\x73\x2f\x6d\x61\x67\x69'
            b'\x63\x2f\x6d\x6f\x78\x65\x6e\x2e\x68\x74\x6d\x6c'
        )

        self.assert_encode_decode(bar, 'GetRequest', decoded, encoded)

        # Message 2.
        decoded = {
            'headerOnly': False,
            'lock': False,
            'url': b'0'
        }

        encoded = b'\x00\x01\x30'

        self.assert_encode_decode(bar, 'GetRequest', decoded, encoded)

    def test_repr_all_types(self):
        all_types = asn1tools.compile_files('tests/files/all_types.asn',
                                            'per')

        self.assertEqual(repr(all_types.types['Boolean']), 'Boolean(Boolean)')
        self.assertEqual(repr(all_types.types['Integer']), 'Integer(Integer)')
        self.assertEqual(repr(all_types.types['Bitstring']), 'BitString(Bitstring)')
        self.assertEqual(repr(all_types.types['Octetstring']), 'OctetString(Octetstring)')
        self.assertEqual(repr(all_types.types['Null']), 'Null(Null)')
        self.assertEqual(repr(all_types.types['Objectidentifier']),
                         'ObjectIdentifier(Objectidentifier)')
        self.assertEqual(repr(all_types.types['Enumerated']), 'Enumerated(Enumerated)')
        self.assertEqual(repr(all_types.types['Utf8string']), 'UTF8String(Utf8string)')
        self.assertEqual(repr(all_types.types['Sequence']), 'Sequence(Sequence, [])')
        self.assertEqual(repr(all_types.types['Set']), 'Set(Set, [])')
        self.assertEqual(repr(all_types.types['Sequence2']),
                         'Sequence(Sequence2, [Integer(a)])')
        self.assertEqual(repr(all_types.types['Set2']), 'Set(Set2, [Integer(a)])')
        self.assertEqual(repr(all_types.types['Numericstring']),
                         'NumericString(Numericstring)')
        self.assertEqual(repr(all_types.types['Printablestring']),
                         'PrintableString(Printablestring)')
        self.assertEqual(repr(all_types.types['Ia5string']), 'IA5String(Ia5string)')
        self.assertEqual(repr(all_types.types['Universalstring']),
                         'UniversalString(Universalstring)')
        self.assertEqual(repr(all_types.types['Visiblestring']),
                         'VisibleString(Visiblestring)')
        self.assertEqual(repr(all_types.types['Generalstring']),
                         'GeneralString(Generalstring)')
        self.assertEqual(repr(all_types.types['Bmpstring']),
                         'BMPString(Bmpstring)')
        self.assertEqual(repr(all_types.types['Teletexstring']),
                         'TeletexString(Teletexstring)')
        self.assertEqual(repr(all_types.types['Utctime']), 'UTCTime(Utctime)')
        self.assertEqual(repr(all_types.types['SequenceOf']),
                         'SequenceOf(SequenceOf, Integer())')
        self.assertEqual(repr(all_types.types['SetOf']), 'SetOf(SetOf, Integer())')

    def test_s1ap_14_4_0(self):
        with self.assertRaises(TypeError):
            s1ap = asn1tools.compile_dict(deepcopy(S1AP_14_4_0), 'per')

            # Message 1.
            decoded_message = (
                'successfulOutcome',
                {
                    'procedureCode': 17,
                    'criticality': 'reject',
                    'value': {
                        'protocolIEs': [
                            {
                                'id': 105,
                                'criticality': 'reject',
                                'value': [
                                    {
                                        'servedPLMNs': [
                                            b'\xab\xcd\xef',
                                            b'\x12\x34\x56'
                                        ],
                                        'servedGroupIDs': [
                                            b'\x22\x22'
                                        ],
                                        'servedMMECs': [
                                            b'\x11'
                                        ]
                                    }
                                ]
                            }
                        ]
                    }
                }
            )

            encoded_message = (
                b'\x20\x11\x00\x15\x00\x00\x01\x00\x69\x00\x0e\x00\x40\xab\xcd\xef\x12'
                b'\x34\x56\x00\x00\x22\x22\x00\x11'
            )

            encoded = s1ap.encode('S1AP-PDU', decoded_message)
            self.assertEqual(encoded, encoded_message)

    def test_information_object(self):
        information_object = asn1tools.compile_files(
            'tests/files/information_object.asn', 'per')

        # Message 1 - without constraints.
        decoded_message = {
            'id': 0,
            'value': b'\x05',
            'comment': 'item 0',
            'extra': 2
        }

        encoded_message = (
            b'\x01\x00\x01\x05\x06\x69\x74\x65\x6d\x20\x30\x01\x02'
        )

        encoded = information_object.encode('ItemWithoutConstraints',
                                            decoded_message)
        self.assertEqual(encoded, encoded_message)

        # Message 1 - with constraints.
        decoded_message = {
            'id': 0,
            'value': True,
            'comment': 'item 0',
            'extra': 2
        }

        encoded_message = (
            b'\x01\x00\x01\x80\x06\x69\x74\x65\x6d\x20\x30\x01\x02'
        )

        # ToDo: Constraints are not yet implemented.
        with self.assertRaises(TypeError) as cm:
            encoded = information_object.encode('ItemWithConstraints',
                                                decoded_message)
            self.assertEqual(encoded, encoded_message)

        self.assertEqual(str(cm.exception), "object of type 'bool' has no len()")

        # Message 2.
        decoded_message = {
            'id': 1,
            'value': {
                'myValue': 7,
                'myType': 0
            },
            'comment': 'item 1',
            'extra': 5
        }

        encoded_message = (
            b'\x01\x01\x05\x02\x01\x07\x01\x00\x06\x69\x74\x65\x6d\x20\x31\x01\x05'
        )

        # ToDo: Constraints are not yet implemented.
        with self.assertRaises(TypeError) as cm:
            encoded = information_object.encode('ItemWithConstraints',
                                                decoded_message)
            self.assertEqual(encoded, encoded_message)

        self.assertTrue("can't concat" in str(cm.exception))

        # Message 3 - error class.
        decoded_message = {
            'errorCategory': 'A',
            'errors': [
                {
                    'errorCode': 1,
                    'errorInfo': 3
                },
                {
                    'errorCode': 2,
                    'errorInfo': True
                }
            ]
        }

        encoded_message = (
            b'\x41\x02\x01\x01\x02\x01\x03\x01\x02\x01\x80'
        )

        # ToDo: Constraints are not yet implemented.
        with self.assertRaises(NotImplementedError):
            encoded = information_object.encode('ErrorReturn', decoded_message)
            self.assertEqual(encoded, encoded_message)

    def test_boolean(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= BOOLEAN "
            "END",
            'per')

        datas = [
            ('A',                     True, b'\x80'),
            ('A',                    False, b'\x00')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(foo, type_name, decoded, encoded)

    def test_integer(self):
        foo = asn1tools.compile_string(
            "Foo DEFINITIONS AUTOMATIC TAGS ::= "
            "BEGIN "
            "A ::= INTEGER "
            "B ::= INTEGER (5..99) "
            "END",
            'per')

        datas = [
            ('A',                    32768, b'\x03\x00\x80\x00'),
            ('A',                    32767, b'\x02\x7f\xff'),
            ('A',                      256, b'\x02\x01\x00'),
            ('A',                      255, b'\x02\x00\xff'),
            ('A',                      128, b'\x02\x00\x80'),
            ('A',                      127, b'\x01\x7f'),
            ('A',                        2, b'\x01\x02'),
            ('A',                        1, b'\x01\x01'),
            ('A',                        0, b'\x01\x00'),
            ('A',                       -1, b'\x01\xff'),
            ('A',                     -128, b'\x01\x80'),
            ('A',                     -129, b'\x02\xff\x7f'),
            ('A',                     -256, b'\x02\xff\x00'),
            ('A',                   -32768, b'\x02\x80\x00'),
            ('A',                   -32769, b'\x03\xff\x7f\xff'),
            ('B',                        5, b'\x00'),
            ('B',                        6, b'\x02'),
            ('B',                       99, b'\xbc')
        ]

        for type_name, decoded, encoded in datas:
            self.assert_encode_decode(foo, type_name, decoded, encoded)


if __name__ == '__main__':
    unittest.main()



"""Basic Encoding Rules (BER) codec.

"""

import math
import binascii
from copy import copy

from . import EncodeError
from . import DecodeError
from . import DecodeTagError
from . import DecodeContentsLengthError
from . import compiler
from .compiler import enum_values_as_dict


class Class(object):
    UNIVERSAL        = 0x00
    APPLICATION      = 0x40
    CONTEXT_SPECIFIC = 0x80
    PRIVATE          = 0xc0


class Encoding(object):
    PRIMITIVE   = 0x00
    CONSTRUCTED = 0x20


class Tag(object):
    END_OF_CONTENTS   = 0x00
    BOOLEAN           = 0x01
    INTEGER           = 0x02
    BIT_STRING        = 0x03
    OCTET_STRING      = 0x04
    NULL              = 0x05
    OBJECT_IDENTIFIER = 0x06
    OBJECT_DESCRIPTOR = 0x07
    EXTERNAL          = 0x08
    REAL              = 0x09
    ENUMERATED        = 0x0a
    EMBEDDED_PDV      = 0x0b
    UTF8_STRING       = 0x0c
    RELATIVE_OID      = 0x0d
    SEQUENCE          = 0x10
    SET               = 0x11
    NUMERIC_STRING    = 0x12
    PRINTABLE_STRING  = 0x13
    T61_STRING        = 0x14
    VIDEOTEX_STRING   = 0x15
    IA5_STRING        = 0x16
    UTC_TIME          = 0x17
    GENERALIZED_TIME  = 0x18
    GRAPHIC_STRING    = 0x19
    VISIBLE_STRING    = 0x1a
    GENERAL_STRING    = 0x1b
    UNIVERSAL_STRING  = 0x1c
    CHARACTER_STRING  = 0x1d
    BMP_STRING        = 0x1e


class DecodeChoiceError(Exception):
    pass


def encode_length_definite(length):
    if length <= 127:
        encoded = bytearray([length])
    else:
        encoded = bytearray()

        while length > 0:
            encoded.append(length & 0xff)
            length >>= 8

        encoded.append(0x80 | len(encoded))
        encoded.reverse()

    return encoded


def decode_length_definite(encoded, offset):
    length = encoded[offset]
    offset += 1

    if length > 127:
        if length == 128:
            raise DecodeError(
                'expected definite length at offset {}, but got indefinite'.format(
                    offset - 1))

        number_of_bytes = (length & 0x7f)
        encoded_length = encoded[offset:number_of_bytes + offset]

        if len(encoded_length) != number_of_bytes:
            raise IndexError(
                'expected {} length byte(s) at offset {}, but got {}'.format(
                    number_of_bytes,
                    offset,
                    len(encoded_length)))

        length = decode_integer(encoded_length)
        offset += number_of_bytes

    if offset + length > len(encoded):
        raise DecodeContentsLengthError(length, offset, len(encoded))

    return length, offset


def decode_length_constructed(encoded, offset):
    length = encoded[offset]

    if length == 128:
        return None, offset + 1
    else:
        return decode_length_definite(encoded, offset)


def decode_integer(data):
    value = 0

    for byte in data:
        value <<= 8
        value += byte

    return value


def encode_signed_integer(data):
    encoded = bytearray()

    if data < 0:
        data *= -1
        data -= 1
        carry = not data

        while data > 0:
            encoded.append((data & 0xff) ^ 0xff)
            carry = (data & 0x80)
            data >>= 8

        if carry:
            encoded.append(0xff)
    elif data > 0:
        while data > 0:
            encoded.append(data & 0xff)
            data >>= 8

        if encoded[-1] & 0x80:
            encoded.append(0)
    else:
        encoded.append(0)

    encoded.append(len(encoded))
    encoded.reverse()

    return encoded


def decode_signed_integer(data):
    value = 0
    is_negative = (data[0] & 0x80)

    for byte in data:
        value <<= 8
        value += byte

    if is_negative:
        value -= (1 << (8 * len(data)))

    return value


def encode_tag(number, flags):
    if number < 31:
        tag = bytearray([flags | number])
    else:
        tag = bytearray([flags | 0x1f])
        encoded = bytearray()

        while number > 0:
            encoded.append(0x80 | (number & 0x7f))
            number >>= 7

        encoded[0] &= 0x7f
        encoded.reverse()
        tag.extend(encoded)

    return tag


def decode_tag(_, offset):
    return 0, 0, offset + 1


def encode_real(data):
    if data == float('inf'):
        data = b'\x40'
    elif data == float('-inf'):
        data = b'\x41'
    elif math.isnan(data):
        data = b'\x42'
    elif data == 0.0:
        data = b''
    else:
        mantissa, exponent = math.frexp(data)
        mantissa = int(mantissa * 2 ** 53)
        lowest_set_bit = (mantissa & -mantissa).bit_length() - 1
        mantissa >>= lowest_set_bit
        mantissa |= (0x80 << (8 * ((mantissa.bit_length() // 8) + 1)))
        mantissa = binascii.unhexlify(hex(mantissa)[4:].rstrip('L'))
        exponent = (52 - lowest_set_bit - exponent)

        if -129 < exponent < 128:
            exponent = [0x80, ((0xff - exponent) & 0xff)]
        elif -32769 < exponent < 32768:
            exponent = ((0xffff - exponent) & 0xffff)
            exponent = [0x81, (exponent >> 8), exponent & 0xff]
        else:
            raise NotImplementedError(
                'REAL exponent {} out of range.'.format(exponent))

        data = bytearray(exponent) + mantissa

    return data


def decode_real(data):
    offset = 0

    if len(data) == 0:
        decoded = 0.0
    else:
        control = data[offset]

        if len(data) == 1:
            try:
                decoded = {
                    0x40: float('inf'),
                    0x41: float('-inf'),
                    0x42: float('nan')
                }[control]
            except KeyError:
                raise NotImplementedError(
                    'Unsupported REAL control word {}.'.format(control))
        else:
            if control == 0x80:
                exponent = data[offset + 1]

                if exponent & 0x80:
                    exponent -= 0x100

                offset += 2
            elif control == 0x81:
                exponent = ((data[offset + 1] << 8) | data[offset + 2])

                if exponent & 0x8000:
                    exponent -= 0x10000

                offset += 3
            else:
                raise NotImplementedError(
                    'Unsupported REAL control word {}.'.format(control))

            mantissa = 0

            for value in data[offset:]:
                mantissa <<= 8
                mantissa += value

            decoded = float(mantissa * 2 ** exponent)

    return decoded


class Type(object):

    def __init__(self, name, type_name, number, flags=0):
        self.name = name
        self.type_name = type_name

        if number is None:
            self.tag = None
        else:
            self.tag = encode_tag(number, flags)

        self.optional = False
        self.default = None

    def set_tag(self, number, flags):
        if not Class.APPLICATION & flags:
            flags |= Class.CONTEXT_SPECIFIC

        self.tag = encode_tag(number, flags)

    def decode_tag(self, data, offset):
        end_offset = offset + len(self.tag)

        if data[offset:end_offset] != self.tag:
            raise DecodeTagError(self.type_name,
                                 self.tag,
                                 data[offset:end_offset],
                                 offset)

        return end_offset


class PrimitiveOrConstructedType(object):

    def __init__(self, name, type_name, number, segment, flags=0):
        self.name = name
        self.type_name = type_name
        self.segment = segment

        if number is None:
            self.tag = None
            self.constructed_tag = None
        else:
            self.tag = encode_tag(number, flags)
            self.constructed_tag = copy(self.tag)
            self.constructed_tag[0] |= Encoding.CONSTRUCTED

        self.optional = False
        self.default = None

    def set_tag(self, number, flags):
        if not Class.APPLICATION & flags:
            flags |= Class.CONTEXT_SPECIFIC

        self.tag = encode_tag(number, flags)
        self.constructed_tag = copy(self.tag)
        self.constructed_tag[0] |= Encoding.CONSTRUCTED

    def decode_tag(self, data, offset):
        end_offset = offset + len(self.tag)
        tag = data[offset:end_offset]

        if tag == self.tag:
            return True, end_offset
        elif tag == self.constructed_tag:
            return False, end_offset
        else:
            raise DecodeTagError(self.type_name,
                                 self.tag,
                                 data[offset:end_offset],
                                 offset)

    def decode(self, data, offset):
        is_primitive, offset = self.decode_tag(data, offset)

        if is_primitive:
            length, offset = decode_length_definite(data, offset)
            end_offset = offset + length

            return self.decode_primitive_contents(data, offset, length), end_offset
        else:
            length, offset = decode_length_constructed(data, offset)
            segments = []

            if length is None:
                while data[offset:offset + 2] != b'\x00\x00':
                    decoded, offset = self.segment.decode(data, offset)
                    segments.append(decoded)

                end_offset = offset + 2
            else:
                end_offset = offset + length

                while offset < end_offset:
                    decoded, offset = self.segment.decode(data, offset)
                    segments.append(decoded)

            return self.decode_constructed_segments(segments), end_offset

    def decode_primitive_contents(self, data, offset, length):
        raise NotImplementedError()

    def decode_constructed_segments(self, segments):
        raise NotImplementedError()


class Integer(Type):

    def __init__(self, name):
        super(Integer, self).__init__(name,
                                      'INTEGER',
                                      Tag.INTEGER)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_signed_integer(data))

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return decode_signed_integer(data[offset:end_offset]), end_offset

    def __repr__(self):
        return 'Integer({})'.format(self.name)


class Real(Type):

    def __init__(self, name):
        super(Real, self).__init__(name, 'REAL', Tag.REAL)

    def encode(self, data, encoded):
        data = encode_real(data)
        encoded.extend(self.tag)
        encoded.append(len(data))
        encoded.extend(data)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length
        decoded = decode_real(data[offset:end_offset])

        return decoded, end_offset

    def __repr__(self):
        return 'Real({})'.format(self.name)


class Boolean(Type):

    def __init__(self, name):
        super(Boolean, self).__init__(name,
                                      'BOOLEAN',
                                      Tag.BOOLEAN)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.append(1)
        encoded.append(0xff * data)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, contents_offset = decode_length_definite(data, offset)

        if length != 1:
            raise DecodeError(
                'expected BOOLEAN contents length 1 at offset {}, but '
                'got {}'.format(offset,
                                length))

        return bool(data[contents_offset]), contents_offset + length

    def __repr__(self):
        return 'Boolean({})'.format(self.name)


class IA5String(PrimitiveOrConstructedType):

    def __init__(self, name):
        super(IA5String, self).__init__(name,
                                        'IA5String',
                                        Tag.IA5_STRING,
                                        OctetString(name))

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data.encode('ascii'))

    def decode_primitive_contents(self, data, offset, length):
        return data[offset:offset + length].decode('ascii')

    def decode_constructed_segments(self, segments):
        return bytearray().join(segments).decode('ascii')

    def __repr__(self):
        return 'IA5String({})'.format(self.name)


class NumericString(PrimitiveOrConstructedType):

    def __init__(self, name):
        super(NumericString, self).__init__(name,
                                            'NumericString',
                                            Tag.NUMERIC_STRING,
                                            OctetString(name))

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data.encode('ascii'))

    def decode_primitive_contents(self, data, offset, length):
        return data[offset:offset + length].decode('ascii')

    def decode_constructed_segments(self, segments):
        return bytearray().join(segments).decode('ascii')

    def __repr__(self):
        return 'NumericString({})'.format(self.name)


class MembersType(Type):

    def __init__(self, name, tag_name, tag, members):
        super(MembersType, self).__init__(name,
                                          tag_name,
                                          tag,
                                          Encoding.CONSTRUCTED)
        self.members = members

    def set_tag(self, number, flags):
        super(MembersType, self).set_tag(number,
                                         flags | Encoding.CONSTRUCTED)

    def encode(self, data, encoded):
        encoded_members = bytearray()

        for member in self.members:
            name = member.name

            if name in data:
                value = data[name]

                if isinstance(member, AnyDefinedBy):
                    member.encode(value, encoded_members, data)
                elif member.default != value or isinstance(member, Null):
                    member.encode(value, encoded_members)
            elif member.optional:
                pass
            elif member.default is None:
                raise EncodeError(
                    "member '{}' not found in {}".format(
                        name,
                        data))

        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(encoded_members)))
        encoded.extend(encoded_members)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)

        if data[offset] == 0x80:
            raise NotImplementedError(
                'decode until an end-of-contents tag is found')
        else:
            length, offset = decode_length_definite(data, offset)

        end_offset = offset + length
        values = {}

        for member in self.members:
            try:
                if offset < end_offset:
                    if isinstance(member, AnyDefinedBy):
                        value, offset = member.decode(data, offset, values)
                    else:
                        value, offset = member.decode(data, offset)
                else:
                    raise IndexError
            except (DecodeError, IndexError) as e:
                if member.optional:
                    continue

                if member.default is None:
                    if isinstance(e, IndexError):
                        e = DecodeError('out of data at offset {}'.format(offset))

                    e.location.append(member.name)
                    raise e

                value = member.default

            values[member.name] = value

        return values, offset

    def __repr__(self):
        return '{}({}, [{}])'.format(
            self.__class__.__name__,
            self.name,
            ', '.join([repr(member) for member in self.members]))


class Sequence(MembersType):

    def __init__(self, name, members):
        super(Sequence, self).__init__(name, 'SEQUENCE', Tag.SEQUENCE, members)


class Set(MembersType):

    def __init__(self, name, members):
        super(Set, self).__init__(name, 'SET', Tag.SET, members)


class ArrayType(Type):

    def __init__(self, name, tag_name, tag, element_type):
        super(ArrayType, self).__init__(name,
                                        tag_name,
                                        tag,
                                        Encoding.CONSTRUCTED)
        self.element_type = element_type

    def set_tag(self, number, flags):
        super(ArrayType, self).set_tag(number,
                                       flags | Encoding.CONSTRUCTED)

    def encode(self, data, encoded):
        encoded_elements = bytearray()

        for entry in data:
            self.element_type.encode(entry, encoded_elements)

        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(encoded_elements)))
        encoded.extend(encoded_elements)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)

        if data[offset] == 0x80:
            raise NotImplementedError(
                'decode until an end-of-contents tag is found')
        else:
            length, offset = decode_length_definite(data, offset)

        decoded = []
        start_offset = offset

        while (offset - start_offset) < length:
            decoded_element, offset = self.element_type.decode(data, offset)
            decoded.append(decoded_element)

        return decoded, offset

    def __repr__(self):
        return '{}({}, {})'.format(self.__class__.__name__,
                                   self.name,
                                   self.element_type)


class SequenceOf(ArrayType):

    def __init__(self, name, element_type):
        super(SequenceOf, self).__init__(name,
                                         'SEQUENCE OF',
                                         Tag.SEQUENCE,
                                         element_type)


class SetOf(ArrayType):

    def __init__(self, name, element_type):
        super(SetOf, self).__init__(name,
                                    'SET OF',
                                    Tag.SET,
                                    element_type)


class BitString(PrimitiveOrConstructedType):

    def __init__(self, name):
        super(BitString, self).__init__(name,
                                        'BIT STRING',
                                        Tag.BIT_STRING,
                                        self)

    def encode(self, data, encoded):
        number_of_unused_bits = (8 - (data[1] % 8))

        if number_of_unused_bits == 8:
            number_of_unused_bits = 0

        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data[0]) + 1))
        encoded.append(number_of_unused_bits)
        encoded.extend(data[0])

    def decode_primitive_contents(self, data, offset, length):
        length -= 1
        number_of_bits = 8 * length - data[offset]
        offset += 1

        return (bytearray(data[offset:offset + length]), number_of_bits)

    def decode_constructed_segments(self, segments):
        decoded = bytearray()
        number_of_bits = 0

        for data, length in segments:
            decoded.extend(data)
            number_of_bits += length

        return (decoded, number_of_bits)

    def __repr__(self):
        return 'BitString({})'.format(self.name)


class OctetString(PrimitiveOrConstructedType):

    def __init__(self, name):
        super(OctetString, self).__init__(name,
                                          'OCTET STRING',
                                          Tag.OCTET_STRING,
                                          self)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data)

    def decode_primitive_contents(self, data, offset, length):
        return bytearray(data[offset:offset + length])

    def decode_constructed_segments(self, segments):
        return bytearray().join(segments)

    def __repr__(self):
        return 'OctetString({})'.format(self.name)


class PrintableString(PrimitiveOrConstructedType):

    def __init__(self, name):
        super(PrintableString, self).__init__(name,
                                              'PrintableString',
                                              Tag.PRINTABLE_STRING,
                                              OctetString(name))

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data.encode('ascii'))

    def decode_primitive_contents(self, data, offset, length):
        return data[offset:offset + length].decode('ascii')

    def decode_constructed_segments(self, segments):
        return bytearray().join(segments).decode('ascii')

    def __repr__(self):
        return 'PrintableString({})'.format(self.name)


class UniversalString(PrimitiveOrConstructedType):

    def __init__(self, name):
        super(UniversalString, self).__init__(name,
                                              'UniversalString',
                                              Tag.UNIVERSAL_STRING,
                                              OctetString(name))

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data.encode('ascii'))

    def decode_primitive_contents(self, data, offset, length):
        return data[offset:offset + length].decode('ascii')

    def decode_constructed_segments(self, segments):
        return bytearray().join(segments).decode('ascii')

    def __repr__(self):
        return 'UniversalString({})'.format(self.name)


class VisibleString(PrimitiveOrConstructedType):

    def __init__(self, name):
        super(VisibleString, self).__init__(name,
                                            'VisibleString',
                                            Tag.VISIBLE_STRING,
                                            OctetString(name))

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data.encode('ascii'))

    def decode_primitive_contents(self, data, offset, length):
        return data[offset:offset + length].decode('ascii')

    def decode_constructed_segments(self, segments):
        return bytearray().join(segments).decode('ascii')

    def __repr__(self):
        return 'VisibleString({})'.format(self.name)


class GeneralString(PrimitiveOrConstructedType):

    def __init__(self, name):
        super(GeneralString, self).__init__(name,
                                            'GeneralString',
                                            Tag.GENERAL_STRING,
                                            OctetString(name))

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data.encode('ascii'))

    def decode_primitive_contents(self, data, offset, length):
        return data[offset:offset + length].decode('ascii')

    def decode_constructed_segments(self, segments):
        return bytearray().join(segments).decode('ascii')

    def __repr__(self):
        return 'GeneralString({})'.format(self.name)


class UTF8String(PrimitiveOrConstructedType):

    def __init__(self, name):
        super(UTF8String, self).__init__(name,
                                         'UTF8String',
                                         Tag.UTF8_STRING,
                                         OctetString(name))

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data.encode('utf-8'))

    def decode_primitive_contents(self, data, offset, length):
        return data[offset:offset + length].decode('utf-8')

    def decode_constructed_segments(self, segments):
        return bytearray().join(segments).decode('utf-8')

    def __repr__(self):
        return 'UTF8String({})'.format(self.name)


class BMPString(PrimitiveOrConstructedType):

    def __init__(self, name):
        super(BMPString, self).__init__(name,
                                        'BMPString',
                                        Tag.BMP_STRING,
                                        OctetString(name))

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data)

    def decode_primitive_contents(self, data, offset, length):
        return bytearray(data[offset:offset + length])

    def decode_constructed_segments(self, segments):
        return bytearray().join(segments)

    def __repr__(self):
        return 'BMPString({})'.format(self.name)


class GraphicString(PrimitiveOrConstructedType):

    def __init__(self, name):
        super(GraphicString, self).__init__(name,
                                            'GraphicString',
                                            Tag.GRAPHIC_STRING,
                                            OctetString(name))

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data.encode('latin-1'))

    def decode_primitive_contents(self, data, offset, length):
        return data[offset:offset + length].decode('latin-1')

    def decode_constructed_segments(self, segments):
        return bytearray().join(segments).decode('latin-1')

    def __repr__(self):
        return 'GraphicString({})'.format(self.name)


class UTCTime(Type):

    def __init__(self, name):
        super(UTCTime, self).__init__(name,
                                      'UTCTime',
                                      Tag.UTC_TIME)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.append(13)
        encoded.extend(bytearray((data).encode('ascii')))

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return data[offset:end_offset].decode('ascii'), end_offset

    def __repr__(self):
        return 'UTCTime({})'.format(self.name)


class GeneralizedTime(Type):

    def __init__(self, name):
        super(GeneralizedTime, self).__init__(name,
                                              'GeneralizedTime',
                                              Tag.GENERALIZED_TIME)

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.append(len(data))
        encoded.extend(data.encode('ascii'))

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return data[offset:end_offset].decode('ascii'), end_offset

    def __repr__(self):
        return 'GeneralizedTime({})'.format(self.name)


class TeletexString(PrimitiveOrConstructedType):

    def __init__(self, name):
        super(TeletexString, self).__init__(name,
                                            'TeletexString',
                                            Tag.T61_STRING,
                                            OctetString(name))

    def encode(self, data, encoded):
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(data)))
        encoded.extend(data)

    def decode_primitive_contents(self, data, offset, length):
        return bytearray(data[offset:offset + length])

    def decode_constructed_segments(self, segments):
        return bytearray().join(segments)

    def __repr__(self):
        return 'TeletexString({})'.format(self.name)


class ObjectIdentifier(Type):

    def __init__(self, name):
        super(ObjectIdentifier, self).__init__(name,
                                               'OBJECT IDENTIFIER',
                                               Tag.OBJECT_IDENTIFIER)

    def encode(self, data, encoded):
        identifiers = [int(identifier) for identifier in data.split('.')]

        first_subidentifier = (40 * identifiers[0] + identifiers[1])
        encoded_subidentifiers = self.encode_subidentifier(
            first_subidentifier)

        for identifier in identifiers[2:]:
            encoded_subidentifiers += self.encode_subidentifier(identifier)

        encoded.extend(self.tag)
        encoded.append(len(encoded_subidentifiers))
        encoded.extend(encoded_subidentifiers)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        subidentifier, offset = self.decode_subidentifier(data, offset)
        decoded = [subidentifier // 40, subidentifier % 40]

        while offset < end_offset:
            subidentifier, offset = self.decode_subidentifier(data, offset)
            decoded.append(subidentifier)

        return '.'.join([str(v) for v in decoded]), end_offset

    def encode_subidentifier(self, subidentifier):
        encoded = [subidentifier & 0x7f]
        subidentifier >>= 7

        while subidentifier > 0:
            encoded.append(0x80 | (subidentifier & 0x7f))
            subidentifier >>= 7

        return encoded[::-1]

    def decode_subidentifier(self, data, offset):
        decoded = 0

        while data[offset] & 0x80:
            decoded += (data[offset] & 0x7f)
            decoded <<= 7
            offset += 1

        decoded += data[offset]

        return decoded, offset + 1

    def __repr__(self):
        return 'ObjectIdentifier({})'.format(self.name)


class Choice(Type):

    def __init__(self, name, members):
        super(Choice, self).__init__(name, 'CHOICE', None)
        self.members = members

    def set_tag(self, number, flags):
        super(Choice, self).set_tag(number,
                                    flags | Encoding.CONSTRUCTED)

    def encode(self, data, encoded):
        if not isinstance(data, tuple):
            raise EncodeError("expected tuple, but got '{}'".format(data))

        for member in self.members:
            if member.name == data[0]:
                member.encode(data[1], encoded)

                return

        raise EncodeError(
            "expected choices are {}, but got '{}'".format(
                [member.name for member in self.members],
                data[0]))

    def decode(self, data, offset):
        for member in self.members:
            if (isinstance(member, Choice)
                or member.tag == data[offset:offset + len(member.tag)]):
                try:
                    decoded, offset = member.decode(data, offset)
                except DecodeChoiceError:
                    pass
                else:
                    return (member.name, decoded), offset

        raise DecodeChoiceError()

    def __repr__(self):
        return 'Choice({}, [{}])'.format(
            self.name,
            ', '.join([repr(member) for member in self.members]))


class Null(Type):

    def __init__(self, name):
        super(Null, self).__init__(name, 'NULL', Tag.NULL)

    def encode(self, _, encoded):
        encoded.extend(self.tag)
        encoded.append(0)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)

        return None, offset + 1

    def __repr__(self):
        return 'Null({})'.format(self.name)


class Any(Type):

    def __init__(self, name):
        super(Any, self).__init__(name, 'ANY', None)

    def encode(self, data, encoded):
        encoded.extend(data)

    def decode(self, data, offset):
        start = offset
        _, _, offset = decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length

        return data[start:end_offset], end_offset

    def __repr__(self):
        return 'Any({})'.format(self.name)


class AnyDefinedBy(Type):

    def __init__(self, name, type_member, choices):
        super(AnyDefinedBy, self).__init__(name,
                                           'ANY DEFINED BY',
                                           None,
                                           None)
        self.type_member = type_member
        self.choices = choices

    def encode(self, data, encoded, values):
        if self.choices:
            try:
                self.choices[values[self.type_member]].encode(data, encoded)
            except KeyError:
                raise EncodeError('bad AnyDefinedBy choice {}'.format(
                    values[self.type_member]))
        else:
            encoded.extend(data)

    def decode(self, data, offset, values):
        if self.choices:
            try:
                return self.choices[values[self.type_member]].decode(data,
                                                                     offset)
            except KeyError:
                raise DecodeError('bad AnyDefinedBy choice {}'.format(
                    values[self.type_member]))
        else:
            start = offset
            _, _, offset = decode_tag(data, offset)
            length, offset = decode_length_definite(data, offset)
            end_offset = offset + length

            return data[start:end_offset], end_offset

    def __repr__(self):
        return 'AnyDefinedBy({})'.format(self.name)


class Enumerated(Type):

    def __init__(self, name, values):
        super(Enumerated, self).__init__(name,
                                         'ENUMERATED',
                                         Tag.ENUMERATED)
        self.values = enum_values_as_dict(values)

    def encode(self, data, encoded):
        for value, name in self.values.items():
            if data == name:
                encoded.extend(self.tag)
                encoded.extend(encode_signed_integer(value))
                return

        raise EncodeError(
            "enumeration value '{}' not found in {}".format(
                data,
                [value for value in self.values.values()]))

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        length, offset = decode_length_definite(data, offset)
        end_offset = offset + length
        value = decode_signed_integer(data[offset:end_offset])

        return self.values[value], end_offset

    def __repr__(self):
        return 'Enumerated({})'.format(self.name)


class ExplicitTag(Type):

    def __init__(self, name, inner):
        super(ExplicitTag, self).__init__(name, 'Tag', None)
        self.inner = inner

    def set_tag(self, number, flags):
        super(ExplicitTag, self).set_tag(number,
                                         flags | Encoding.CONSTRUCTED)

    def encode(self, data, encoded):
        encoded_inner = bytearray()
        self.inner.encode(data, encoded_inner)
        encoded.extend(self.tag)
        encoded.extend(encode_length_definite(len(encoded_inner)))
        encoded.extend(encoded_inner)

    def decode(self, data, offset):
        offset = self.decode_tag(data, offset)
        _, offset = decode_length_definite(data, offset)
        return self.inner.decode(data, offset)

    def __repr__(self):
        return 'Tag()'


class Recursive(Type):

    def __init__(self, name, type_name, module_name):
        super(Recursive, self).__init__(name, 'RECURSIVE', None)
        self.type_name = type_name
        self.module_name = module_name

    def encode(self, _data, _encoded):
        raise NotImplementedError(
            "Recursive types are not yet implemented (type '{}').".format(
                self.type_name))

    def decode(self, _data, _offset):
        raise NotImplementedError(
            "Recursive types are not yet implemented (type '{}').".format(
                self.type_name))

    def __repr__(self):
        return 'Recursive({})'.format(self.name)


class CompiledType(object):

    def __init__(self, type_):
        self._type = type_

    def encode(self, data):
        encoded = bytearray()
        self._type.encode(data, encoded)

        return encoded

    def decode(self, data):
        return self._type.decode(bytearray(data), 0)[0]

    def __repr__(self):
        return repr(self._type)


class Compiler(compiler.Compiler):

    def process_type(self, type_name, type_descriptor, module_name):
        return CompiledType(self.compile_type(type_name,
                                              type_descriptor,
                                              module_name))

    def compile_implicit_type(self, name, type_descriptor, module_name):
        type_name = type_descriptor['type']

        if type_name == 'SEQUENCE':
            compiled = Sequence(
                name,
                self.compile_members(type_descriptor['members'],
                                     module_name))
        elif type_name == 'SEQUENCE OF':
            compiled = SequenceOf(name,
                                  self.compile_type('',
                                                    type_descriptor['element'],
                                                    module_name))
        elif type_name == 'SET':
            compiled = Set(
                name,
                self.compile_members(type_descriptor['members'],
                                     module_name))
        elif type_name == 'SET OF':
            compiled = SetOf(name,
                             self.compile_type('',
                                               type_descriptor['element'],
                                               module_name))
        elif type_name == 'CHOICE':
            compiled = Choice(
                name,
                self.compile_members(type_descriptor['members'],
                                     module_name))
        elif type_name == 'INTEGER':
            compiled = Integer(name)
        elif type_name == 'REAL':
            compiled = Real(name)
        elif type_name == 'ENUMERATED':
            compiled = Enumerated(name, type_descriptor['values'])
        elif type_name == 'BOOLEAN':
            compiled = Boolean(name)
        elif type_name == 'OBJECT IDENTIFIER':
            compiled = ObjectIdentifier(name)
        elif type_name == 'OCTET STRING':
            compiled = OctetString(name)
        elif type_name == 'TeletexString':
            compiled = TeletexString(name)
        elif type_name == 'NumericString':
            compiled = NumericString(name)
        elif type_name == 'PrintableString':
            compiled = PrintableString(name)
        elif type_name == 'IA5String':
            compiled = IA5String(name)
        elif type_name == 'VisibleString':
            compiled = VisibleString(name)
        elif type_name == 'GeneralString':
            compiled = GeneralString(name)
        elif type_name == 'UTF8String':
            compiled = UTF8String(name)
        elif type_name == 'BMPString':
            compiled = BMPString(name)
        elif type_name == 'GraphicString':
            compiled = GraphicString(name)
        elif type_name == 'UTCTime':
            compiled = UTCTime(name)
        elif type_name == 'UniversalString':
            compiled = UniversalString(name)
        elif type_name == 'GeneralizedTime':
            compiled = GeneralizedTime(name)
        elif type_name == 'BIT STRING':
            compiled = BitString(name)
        elif type_name == 'ANY':
            compiled = Any(name)
        elif type_name == 'ANY DEFINED BY':
            choices = {}

            for key, value in type_descriptor['choices'].items():
                choices[key] = self.compile_type(key,
                                                 value,
                                                 module_name)

            compiled = AnyDefinedBy(name,
                                    type_descriptor['value'],
                                    choices)
        elif type_name == 'NULL':
            compiled = Null(name)
        else:
            if type_name in self.types_backtrace:
                compiled = Recursive(name,
                                     type_name,
                                     module_name)
            else:
                self.types_backtrace_push(type_name)
                compiled = self.compile_type(
                    name,
                    *self.lookup_type_descriptor(
                        type_name,
                        module_name))
                self.types_backtrace_pop()

        return compiled

    def compile_type(self, name, type_descriptor, module_name):
        compiled = self.compile_implicit_type(name,
                                              type_descriptor,
                                              module_name)

        if self.is_explicit_tag(type_descriptor):
            compiled = ExplicitTag(name, compiled)

        # Set any given tag.
        if 'tag' in type_descriptor:
            tag = type_descriptor['tag']
            class_ = tag.get('class', None)

            if class_ == 'APPLICATION':
                flags = Class.APPLICATION
            elif class_ == 'PRIVATE':
                flags = Class.PRIVATE
            else:
                flags = 0

            compiled.set_tag(tag['number'], flags)

        return compiled

    def compile_members(self, members, module_name):
        compiled_members = []

        for member in members:
            if member == '...':
                continue

            if isinstance(member, list):
                compiled_members.extend(self.compile_members(member,
                                                             module_name))
                continue

            compiled_member = self.compile_type(member['name'],
                                                member,
                                                module_name)

            if 'optional' in member:
                compiled_member.optional = member['optional']

            if 'default' in member:
                compiled_member.default = member['default']

            compiled_members.append(compiled_member)

        return compiled_members


def compile_dict(specification):
    return Compiler(specification).process()


def decode_length(data):
    try:
        return sum(decode_length_definite(bytearray(data), 1))
    except DecodeContentsLengthError as e:
        return (e.length + e.offset)
    except IndexError:
        return None



import json
import unittest
import asn1tools
import sys
import math
from copy import deepcopy

sys.path.append('tests/files')
sys.path.append('tests/files/3gpp')

from rrc_8_6_0 import EXPECTED as RRC_8_6_0


def loadb(encoded):
    return json.loads(encoded.decode('utf-8'))


class Asn1ToolsJerTest(unittest.TestCase):

    maxDiff = None

    def test_foo(self):
        foo = asn1tools.compile_files(['tests/files/foo.asn'], 'jer')

        self.assertEqual(len(foo.types), 2)
        self.assertTrue(foo.types['Question'] is not None)
        self.assertTrue(foo.types['Answer'] is not None)
        self.assertEqual(len(foo.modules), 1)
        self.assertTrue(foo.modules['Foo'] is not None)

        # Question.
        decoded_message = {'id': 1, 'question': 'Is 1+1=3?'}
        encoded_message = b'{"id":1,"question":"Is 1+1=3?"}'

        encoded = foo.encode('Question', decoded_message)
        self.assertEqual(loadb(encoded), loadb(encoded_message))
        decoded = foo.decode('Question', encoded)
        self.assertEqual(decoded, decoded_message)

        # Answer.
        decoded_message = {'id': 1, 'answer': False}
        encoded_message = b'{"id":1,"answer":false}'

        encoded = foo.encode('Answer', decoded_message)
        self.assertEqual(loadb(encoded), loadb(encoded_message))
        decoded = foo.decode('Answer', encoded)
        self.assertEqual(decoded, decoded_message)

        # Encode a question with missing field 'id'.
        with self.assertRaises(asn1tools.EncodeError) as cm:
            encoded = foo.encode('Question', {'question': 'Is 1+1=3?'})

        self.assertEqual(
            str(cm.exception),
            "Sequence member 'id' not found in {'question': 'Is 1+1=3?'}.")

    def test_decode_length(self):
        foo = asn1tools.compile_files('tests/files/foo.asn', 'jer')

        with self.assertRaises(asn1tools.DecodeError) as cm:
            foo.decode_length(b'')

        self.assertEqual(str(cm.exception),
                         ': Decode length is not supported for this codec.')

    def test_rrc_8_6_0(self):
        rrc = asn1tools.compile_dict(deepcopy(RRC_8_6_0), 'jer')

        # Message 1.
        decoded_message = {
            'message': (
                'c1',
                (
                    'paging',
                    {
                        'systemInfoModification': 'true',
                        'nonCriticalExtension': {
                        }
                    }
                )
            )
        }

        encoded_message = (
            b'{"message":{"c1":{"paging":{"systemInfoModification":"true","'
            b'nonCriticalExtension":{}}}}}'
        )

        encoded = rrc.encode('PCCH-Message', decoded_message)
        self.assertEqual(loadb(encoded), loadb(encoded_message))
        decoded = rrc.decode('PCCH-Message', encoded)
        self.assertEqual(decoded, decoded_message)

        # Message 2.
        decoded_message = {
            'message': {
                'dl-Bandwidth': 'n6',
                'phich-Config': {
                    'phich-Duration': 'normal',
                    'phich-Resource': 'half'
                },
                'systemFrameNumber': (b'\x12', 8),
                'spare': (b'\x34\x40', 10)
            }
        }

        encoded_message = (
            b'{"message":{"dl-Bandwidth":"n6","phich-Config":{"phich-Duration'
            b'":"normal","phich-Resource":"half"},"spare":"3440","'
            b'systemFrameNumber":"12"}}'
        )

        encoded = rrc.encode('BCCH-BCH-Message', decoded_message)
        self.assertEqual(loadb(encoded), loadb(encoded_message))
        decoded = rrc.decode('BCCH-BCH-Message', encoded)
        self.assertEqual(decoded, decoded_message)

        # Message 3.
        decoded_message = {
            'message': (
                'c1',
                (
                    'systemInformation',
                    {
                        'criticalExtensions': (
                            'systemInformation-r8',
                            {
                                'sib-TypeAndInfo': [
                                    (
                                        'sib2',
                                        {
                                            'ac-BarringInfo': {
                                                'ac-BarringForEmergency': True,
                                                'ac-BarringForMO-Data': {
                                                    'ac-BarringFactor': 'p95',
                                                    'ac-BarringTime': 's128',
                                                    'ac-BarringForSpecialAC': (b'\xf0', 5)
                                                }
                                            },
                                            'radioResourceConfigCommon': {
                                                'rach-ConfigCommon': {
                                                    'preambleInfo': {
                                                        'numberOfRA-Preambles': 'n24',
                                                        'preamblesGroupAConfig': {
                                                            'sizeOfRA-PreamblesGroupA': 'n28',
                                                            'messageSizeGroupA': 'b144',
                                                            'messagePowerOffsetGroupB': 'minusinfinity'
                                                        }
                                                    },
                                                    'powerRampingParameters': {
                                                        'powerRampingStep': 'dB0',
                                                        'preambleInitialReceivedTargetPower': 'dBm-102'
                                                    },
                                                    'ra-SupervisionInfo': {
                                                        'preambleTransMax': 'n8',
                                                        'ra-ResponseWindowSize': 'sf6',
                                                        'mac-ContentionResolutionTimer': 'sf48'
                                                    },
                                                    'maxHARQ-Msg3Tx': 8
                                                },
                                                'bcch-Config': {
                                                    'modificationPeriodCoeff': 'n2'
                                                },
                                                'pcch-Config': {
                                                    'defaultPagingCycle': 'rf256',
                                                    'nB': 'twoT'
                                                },
                                                'prach-Config': {
                                                    'rootSequenceIndex': 836,
                                                    'prach-ConfigInfo': {
                                                        'prach-ConfigIndex': 33,
                                                        'highSpeedFlag': False,
                                                        'zeroCorrelationZoneConfig': 10,
                                                        'prach-FreqOffset': 64
                                                    }
                                                },
                                                'pdsch-ConfigCommon': {
                                                    'referenceSignalPower': -60,
                                                    'p-b': 2
                                                },
                                                'pusch-ConfigCommon': {
                                                    'pusch-ConfigBasic': {
                                                        'n-SB': 1,
                                                        'hoppingMode': 'interSubFrame',
                                                        'pusch-HoppingOffset': 10,
                                                        'enable64QAM': False
                                                    },
                                                    'ul-ReferenceSignalsPUSCH': {
                                                        'groupHoppingEnabled': True,
                                                        'groupAssignmentPUSCH': 22,
                                                        'sequenceHoppingEnabled': False,
                                                        'cyclicShift': 5
                                                    }
                                                },
                                                'pucch-ConfigCommon': {
                                                    'deltaPUCCH-Shift': 'ds1',
                                                    'nRB-CQI': 98,
                                                    'nCS-AN': 4,
                                                    'n1PUCCH-AN': 2047
                                                },
                                                'soundingRS-UL-ConfigCommon': (
                                                    'setup',
                                                    {
                                                        'srs-BandwidthConfig': 'bw0',
                                                        'srs-SubframeConfig': 'sc4',
                                                        'ackNackSRS-SimultaneousTransmission': True
                                                    }),
                                                'uplinkPowerControlCommon': {
                                                    'p0-NominalPUSCH': -126,
                                                    'alpha': 'al0',
                                                    'p0-NominalPUCCH': -127,
                                                    'deltaFList-PUCCH': {
                                                        'deltaF-PUCCH-Format1': 'deltaF-2',
                                                        'deltaF-PUCCH-Format1b': 'deltaF1',
                                                        'deltaF-PUCCH-Format2': 'deltaF0',
                                                        'deltaF-PUCCH-Format2a': 'deltaF-2',
                                                        'deltaF-PUCCH-Format2b': 'deltaF0'
                                                    },
                                                    'deltaPreambleMsg3': -1
                                                },
                                                'ul-CyclicPrefixLength': 'len1'
                                            },
                                            'ue-TimersAndConstants': {
                                                't300': 'ms100',
                                                't301': 'ms200',
                                                't310': 'ms50',
                                                'n310': 'n2',
                                                't311': 'ms30000',
                                                'n311': 'n2'
                                            },
                                            'freqInfo': {
                                                'additionalSpectrumEmission': 3
                                            },
                                            'timeAlignmentTimerCommon': 'sf500'
                                        }
                                    ),
                                    (
                                        'sib3',
                                        {
                                            'cellReselectionInfoCommon': {
                                                'q-Hyst': 'dB0',
                                                'speedStateReselectionPars': {
                                                    'mobilityStateParameters': {
                                                        't-Evaluation': 's180',
                                                        't-HystNormal': 's180',
                                                        'n-CellChangeMedium': 1,
                                                        'n-CellChangeHigh': 16
                                                    },
                                                    'q-HystSF': {
                                                        'sf-Medium': 'dB-6',
                                                        'sf-High': 'dB-4'
                                                    }
                                                }
                                            },
                                            'cellReselectionServingFreqInfo': {
                                                'threshServingLow': 7,
                                                'cellReselectionPriority': 3
                                            },
                                            'intraFreqCellReselectionInfo': {
                                                'q-RxLevMin': -33,
                                                's-IntraSearch': 0,
                                                'presenceAntennaPort1': False,
                                                'neighCellConfig': (b'\x80', 2),
                                                't-ReselectionEUTRA': 4
                                            }
                                        }
                                    ),
                                    (
                                        'sib4',
                                        {
                                        }
                                    ),
                                    (
                                        'sib5',
                                        {
                                            'interFreqCarrierFreqList': [
                                                {
                                                    'dl-CarrierFreq': 1,
                                                    'q-RxLevMin': -45,
                                                    't-ReselectionEUTRA': 0,
                                                    'threshX-High': 31,
                                                    'threshX-Low': 29,
                                                    'allowedMeasBandwidth': 'mbw6',
                                                    'presenceAntennaPort1': True,
                                                    'neighCellConfig': (b'\x00', 2),
                                                    'q-OffsetFreq': 'dB0'
                                                }
                                            ]
                                        }
                                    ),
                                    (
                                        'sib6',
                                        {
                                            't-ReselectionUTRA': 3
                                        }
                                    ),
                                    ('sib7',
                                     {
                                         't-ReselectionGERAN': 3
                                     }
                                    ),
                                    ('sib8',
                                     {
                                         'parameters1XRTT': {
                                             'longCodeState1XRTT': (b'\x01#Eg\x89\x00', 42)
                                         }
                                     }
                                    ),
                                    (
                                        'sib9',
                                        {
                                            'hnb-Name': b'4'
                                        }
                                    ),
                                    ('sib10',
                                     {
                                         'messageIdentifier': (b'#4', 16),
                                         'serialNumber': (b'\x124', 16),
                                         'warningType': b'2\x12'
                                     }
                                    ),
                                    (
                                        'sib11',
                                        {
                                            'messageIdentifier': (b'g\x88', 16),
                                            'serialNumber': (b'T5', 16),
                                            'warningMessageSegmentType': 'notLastSegment',
                                            'warningMessageSegmentNumber': 19,
                                            'warningMessageSegment': b'\x12'
                                        }
                                    )
                                ]
                            }
                        )
                    }
                )
            )
        }

        encoded_message = (
            b'{"message":{"c1":{"systemInformation":{"criticalExtensions":{"sy'
            b'stemInformation-r8":{"sib-TypeAndInfo":[{"sib2":{"ac-BarringInfo'
            b'":{"ac-BarringForEmergency":true,"ac-BarringForMO-Data":{"ac-Bar'
            b'ringFactor":"p95","ac-BarringForSpecialAC":"f0","ac-BarringTime"'
            b':"s128"}},"freqInfo":{"additionalSpectrumEmission":3},"radioReso'
            b'urceConfigCommon":{"bcch-Config":{"modificationPeriodCoeff":"n2"'
            b'},"pcch-Config":{"defaultPagingCycle":"rf256","nB":"twoT"},"pdsc'
            b'h-ConfigCommon":{"p-b":2,"referenceSignalPower":-60},"prach-Conf'
            b'ig":{"prach-ConfigInfo":{"highSpeedFlag":false,"prach-ConfigInde'
            b'x":33,"prach-FreqOffset":64,"zeroCorrelationZoneConfig":10},"roo'
            b'tSequenceIndex":836},"pucch-ConfigCommon":{"deltaPUCCH-Shift":"d'
            b's1","n1PUCCH-AN":2047,"nCS-AN":4,"nRB-CQI":98},"pusch-ConfigComm'
            b'on":{"pusch-ConfigBasic":{"enable64QAM":false,"hoppingMode":"int'
            b'erSubFrame","n-SB":1,"pusch-HoppingOffset":10},"ul-ReferenceSign'
            b'alsPUSCH":{"cyclicShift":5,"groupAssignmentPUSCH":22,"groupHoppi'
            b'ngEnabled":true,"sequenceHoppingEnabled":false}},"rach-ConfigCom'
            b'mon":{"maxHARQ-Msg3Tx":8,"powerRampingParameters":{"powerRamping'
            b'Step":"dB0","preambleInitialReceivedTargetPower":"dBm-102"},"pre'
            b'ambleInfo":{"numberOfRA-Preambles":"n24","preamblesGroupAConfig"'
            b':{"messagePowerOffsetGroupB":"minusinfinity","messageSizeGroupA"'
            b':"b144","sizeOfRA-PreamblesGroupA":"n28"}},"ra-SupervisionInfo":'
            b'{"mac-ContentionResolutionTimer":"sf48","preambleTransMax":"n8",'
            b'"ra-ResponseWindowSize":"sf6"}},"soundingRS-UL-ConfigCommon":{"s'
            b'etup":{"ackNackSRS-SimultaneousTransmission":true,"srs-Bandwidth'
            b'Config":"bw0","srs-SubframeConfig":"sc4"}},"ul-CyclicPrefixLengt'
            b'h":"len1","uplinkPowerControlCommon":{"alpha":"al0","deltaFList-'
            b'PUCCH":{"deltaF-PUCCH-Format1":"deltaF-2","deltaF-PUCCH-Format1b'
            b'":"deltaF1","deltaF-PUCCH-Format2":"deltaF0","deltaF-PUCCH-Forma'
            b't2a":"deltaF-2","deltaF-PUCCH-Format2b":"deltaF0"},"deltaPreambl'
            b'eMsg3":-1,"p0-NominalPUCCH":-127,"p0-NominalPUSCH":-126}},"timeA'
            b'lignmentTimerCommon":"sf500","ue-TimersAndConstants":{"n310":"n2'
            b'","n311":"n2","t300":"ms100","t301":"ms200","t310":"ms50","t311"'
            b':"ms30000"}}},{"sib3":{"cellReselectionInfoCommon":{"q-Hyst":"dB'
            b'0","speedStateReselectionPars":{"mobilityStateParameters":{"n-Ce'
            b'llChangeHigh":16,"n-CellChangeMedium":1,"t-Evaluation":"s180","t'
            b'-HystNormal":"s180"},"q-HystSF":{"sf-High":"dB-4","sf-Medium":"d'
            b'B-6"}}},"cellReselectionServingFreqInfo":{"cellReselectionPriori'
            b'ty":3,"threshServingLow":7},"intraFreqCellReselectionInfo":{"nei'
            b'ghCellConfig":"80","presenceAntennaPort1":false,"q-RxLevMin":-33'
            b',"s-IntraSearch":0,"t-ReselectionEUTRA":4}}},{"sib4":{}},{"sib5"'
            b':{"interFreqCarrierFreqList":[{"allowedMeasBandwidth":"mbw6","dl'
            b'-CarrierFreq":1,"neighCellConfig":"00","presenceAntennaPort1":tr'
            b'ue,"q-OffsetFreq":"dB0","q-RxLevMin":-45,"t-ReselectionEUTRA":0,'
            b'"threshX-High":31,"threshX-Low":29}]}},{"sib6":{"t-ReselectionUT'
            b'RA":3}},{"sib7":{"t-ReselectionGERAN":3}},{"sib8":{"parameters1X'
            b'RTT":{"longCodeState1XRTT":"012345678900"}}},{"sib9":{"hnb-Name"'
            b':"34"}},{"sib10":{"messageIdentifier":"2334","serialNumber":"123'
            b'4","warningType":"3212"}},{"sib11":{"messageIdentifier":"6788","'
            b'serialNumber":"5435","warningMessageSegment":"12","warningMessag'
            b'eSegmentNumber":19,"warningMessageSegmentType":"notLastSegment"}'
            b'}]}}}}}}'
        )

        encoded = rrc.encode('BCCH-DL-SCH-Message', decoded_message)
        self.assertEqual(loadb(encoded), loadb(encoded_message))
        decoded = rrc.decode('BCCH-DL-SCH-Message', encoded)
        self.assertEqual(decoded, decoded_message)

    def test_all_types(self):
        all_types = asn1tools.compile_files('tests/files/all_types.asn', 'jer')

        datas = [
            ('Boolean',             True, b'true'),
            ('Integer',              127, b'127'),
            ('Integer',                0, b'0'),
            ('Integer',             -128, b'-128'),
            ('Real',                 1.0, b'1.0'),
            ('Real',                -2.0, b'-2.0'),
            ('Real',        float('inf'), b'"INF"'),
            ('Real',       float('-inf'), b'"-INF"'),
            ('Octetstring',      b'\x00', b'"00"'),
            ('Null',                None, b'null'),
            ('Enumerated',         'one', b'"one"'),
            ('Utf8string',         'foo', b'"foo"'),
            ('Sequence',              {}, b'{}'),
            ('Sequence2',       {'a': 1}, b'{"a":1}'),
            ('Set',                   {}, b'{}'),
            ('Set2',            {'a': 2}, b'{"a":2}'),
            ('Numericstring',      '123', b'"123"'),
            ('Printablestring',    'foo', b'"foo"'),
            ('Ia5string',          'bar', b'"bar"'),
            ('Universalstring',    'bar', b'"bar"'),
            ('Visiblestring',      'bar', b'"bar"'),
            ('Generalstring',      'bar', b'"bar"'),
            ('Bmpstring',         b'bar', b'"bar"'),
            ('Teletexstring',     b'fum', b'"fum"'),
            ('SequenceOf',            [], b'[]'),
            ('SetOf',                 [], b'[]'),
        ]

        for type_name, decoded, encoded in datas:
            self.assertEqual(all_types.encode(type_name, decoded), encoded)
            self.assertEqual(all_types.decode(type_name, encoded), decoded)

        self.assertEqual(all_types.encode('Real', float('nan')), b'"NaN"')
        self.assertTrue(math.isnan(all_types.decode('Real', b'"NaN"')))

        self.assertEqual(all_types.decode('Real', b'"0"'), 0.0)
        self.assertEqual(all_types.decode('Real', b'"-0"'), 0.0)

        decoded_message = (b'\x80', 1)
        encoded_message = b'{"value":"80","length":1}'

        self.assertEqual(loadb(all_types.encode('Bitstring', decoded_message)),
                         loadb(encoded_message))
        self.assertEqual(all_types.decode('Bitstring', encoded_message),
                         decoded_message)

        with self.assertRaises(NotImplementedError):
            all_types.encode('Sequence12', {'a': [{'a': []}]})

        with self.assertRaises(NotImplementedError):
            all_types.decode('Sequence12', b'{"a": [{"a": []}]}')

    def test_repr_all_types(self):
        all_types = asn1tools.compile_files('tests/files/all_types.asn',
                                            'jer')

        self.assertEqual(repr(all_types.types['Boolean']), 'Boolean(Boolean)')
        self.assertEqual(repr(all_types.types['Integer']), 'Integer(Integer)')
        self.assertEqual(repr(all_types.types['Real']), 'Real(Real)')
        self.assertEqual(repr(all_types.types['Bitstring']), 'BitString(Bitstring)')
        self.assertEqual(repr(all_types.types['Octetstring']), 'OctetString(Octetstring)')
        self.assertEqual(repr(all_types.types['Null']), 'Null(Null)')
        self.assertEqual(repr(all_types.types['Objectidentifier']),
                         'ObjectIdentifier(Objectidentifier)')
        self.assertEqual(repr(all_types.types['Enumerated']), 'Enumerated(Enumerated)')
        self.assertEqual(repr(all_types.types['Utf8string']), 'UTF8String(Utf8string)')
        self.assertEqual(repr(all_types.types['Sequence']), 'Sequence(Sequence, [])')
        self.assertEqual(repr(all_types.types['Set']), 'Set(Set, [])')
        self.assertEqual(repr(all_types.types['Sequence2']),
                         'Sequence(Sequence2, [Integer(a)])')
        self.assertEqual(repr(all_types.types['Set2']), 'Set(Set2, [Integer(a)])')
        self.assertEqual(repr(all_types.types['Numericstring']),
                         'NumericString(Numericstring)')
        self.assertEqual(repr(all_types.types['Printablestring']),
                         'PrintableString(Printablestring)')
        self.assertEqual(repr(all_types.types['Ia5string']), 'IA5String(Ia5string)')
        self.assertEqual(repr(all_types.types['Universalstring']),
                         'UniversalString(Universalstring)')
        self.assertEqual(repr(all_types.types['Visiblestring']),
                         'VisibleString(Visiblestring)')
        self.assertEqual(repr(all_types.types['Generalstring']),
                         'GeneralString(Generalstring)')
        self.assertEqual(repr(all_types.types['Bmpstring']),
                         'BMPString(Bmpstring)')
        self.assertEqual(repr(all_types.types['Teletexstring']),
                         'TeletexString(Teletexstring)')
        self.assertEqual(repr(all_types.types['Utctime']), 'UTCTime(Utctime)')
        self.assertEqual(repr(all_types.types['SequenceOf']),
                         'SequenceOf(SequenceOf, Integer())')
        self.assertEqual(repr(all_types.types['SetOf']), 'SetOf(SetOf, Integer())')

    def test_all_types_automatic_tags(self):
        all_types = asn1tools.compile_files(
            'tests/files/all_types_automatic_tags.asn', 'jer')

        datas = [
            ('Sequence3', {'a': 1, 'c': 2,'d': True}, b'{"a":1,"c":2,"d":true}')
        ]

        for type_name, decoded, encoded in datas:
            self.assertEqual(loadb(all_types.encode(type_name, decoded)),
                             loadb(encoded))
            self.assertEqual(all_types.decode(type_name, encoded), decoded)


if __name__ == '__main__':
    unittest.main()



"""Base Compiler class used by all codecs.

"""

from copy import deepcopy
from ..errors import CompileError


def flatten(dlist):
    flist = []

    for item in dlist:
        if isinstance(item, list):
            flist.extend(item)
        else:
            flist.append(item)

    return flist


class Compiler(object):

    def __init__(self, specification):
        self._specification = specification
        self._types_backtrace = []

    def types_backtrace_push(self, type_name):
        self._types_backtrace.append(type_name)

    def types_backtrace_pop(self):
        self._types_backtrace.pop()

    @property
    def types_backtrace(self):
        return self._types_backtrace

    def process(self):
        self.pre_process()

        compiled = {}

        for module_name in self._specification:
            items = self._specification[module_name]['types'].items()

            for type_name, type_descriptor in items:
                self.types_backtrace_push(type_name)
                compiled_type = self.process_type(type_name,
                                                  type_descriptor,
                                                  module_name)
                self.types_backtrace_pop()

                if module_name not in compiled:
                    compiled[module_name] = {}

                compiled[module_name][type_name] = compiled_type

        return compiled

    def pre_process(self):
        for module_name in self._specification:
            module = self._specification[module_name]

            self.pre_process_components_of(module, module_name)

            if module['extensibility-implied']:
                self.pre_process_extensibility_implied(module)

            self.pre_process_tags(module, module_name)

        return self._specification

    def pre_process_components_of(self, module, module_name):
        for type_descriptor in module['types'].values():
            self.pre_process_components_of_type(type_descriptor,
                                                module_name)

    def pre_process_components_of_type(self, type_descriptor, module_name):
        type_name = type_descriptor['type']

        if type_name in ['SEQUENCE', 'SET', 'CHOICE']:
            type_descriptor['members'] = self.pre_process_components_of_expand_members(
                type_descriptor['members'],
                module_name)

    def pre_process_components_of_expand_members(self, members, module_name):
        expanded_members = []

        for member in members:
            if 'components-of' in member:
                type_descriptor, inner_module_name = self.lookup_type_descriptor(
                    member['components-of'],
                    module_name)
                inner_members = self.pre_process_components_of_expand_members(
                    type_descriptor['members'],
                    inner_module_name)

                for inner_member in inner_members:
                    if inner_member == '...':
                        break

                    expanded_members.append(deepcopy(inner_member))
            else:
                expanded_members.append(member)

        return expanded_members

    def pre_process_extensibility_implied(self, module):
        for type_descriptor in module['types'].values():
            self.pre_process_extensibility_implied_type(type_descriptor)

    def pre_process_extensibility_implied_type(self, type_descriptor):
        type_name = type_descriptor['type']

        if type_name in ['SEQUENCE', 'SET', 'CHOICE']:
            members = type_descriptor['members']

            for member in members:
                if member == '...' or isinstance(member, list):
                    continue

                self.pre_process_extensibility_implied_type(member)

            if '...' not in members:
                members.append('...')

    def pre_process_tags(self, module, module_name):
        module_tags = module.get('tags', 'EXPLICIT')

        for type_descriptor in module['types'].values():
            self.pre_process_tags_type(type_descriptor,
                                       module_tags,
                                       module_name)

    def pre_process_tags_type(self,
                              type_descriptor,
                              module_tags,
                              module_name):
        type_name = type_descriptor['type']

        if 'tag' in type_descriptor:
            tag = type_descriptor['tag']

            if 'kind' not in tag:
                if self.resolve_type_name(type_name, module_name) == 'CHOICE':
                    tag['kind'] = 'EXPLICIT'
                elif module_tags in ['IMPLICIT', 'EXPLICIT']:
                    tag['kind'] = module_tags
                else:
                    tag['kind'] = 'IMPLICIT'

        if type_name in ['SEQUENCE', 'SET', 'CHOICE']:
            self.pre_process_tags_type_members(type_descriptor,
                                               module_tags,
                                               module_name)

        if type_name in ['SEQUENCE OF', 'SET OF']:
            self.pre_process_tags_type(type_descriptor['element'],
                                       module_tags,
                                       module_name)

    def pre_process_tags_type_members(self,
                                      type_descriptor,
                                      module_tags,
                                      module_name):
        def is_any_member_tagged(members):
            for member in members:
                if member == '...':
                    continue

                if 'tag' in member:
                    return True

            return False

        number = None
        members = flatten(type_descriptor['members'])

        # Add tag number to all members if AUTOMATIC TAGS are
        # selected and no member is tagged.
        if module_tags == 'AUTOMATIC' and not is_any_member_tagged(members):
            number = 0

        for member in members:
            if member == '...':
                continue

            if number is not None:
                if 'tag' not in member:
                    member['tag'] = {}

                member['tag']['number'] = number
                number += 1

            self.pre_process_tags_type(member,
                                       module_tags,
                                       module_name)

    def resolve_type_name(self, type_name, module_name):
        try:
            while True:
                type_descriptor, module_name = self.lookup_type_descriptor(
                    type_name,
                    module_name)
                type_name = type_descriptor['type']
        except CompileError:
            pass

        return type_name

    def process_type(self, type_name, type_descriptor, module_name):
        return NotImplementedError()

    def get_size_range(self, type_descriptor, module_name):
        """Returns a tuple of the minimum and maximum values allowed according
        the the ASN.1 specification SIZE parameter. Returns (None,
        None, None) if the type does not have a SIZE parameter.

        """

        size = type_descriptor.get('size', None)

        if size is None:
            minimum = None
            maximum = None
            has_extension_marker = None
        else:
            if isinstance(size[0], tuple):
                minimum, maximum = size[0]
            else:
                minimum = size[0]
                maximum = size[0]

            has_extension_marker = (size[-1] == '...')

        if isinstance(minimum, str):
            minimum = self.lookup_value(minimum, module_name)[0]['value']

        if isinstance(maximum, str):
            maximum = self.lookup_value(maximum, module_name)[0]['value']

        return minimum, maximum, has_extension_marker

    def get_restricted_to_range(self, type_descriptor, module_name):
        restricted_to = type_descriptor.get('restricted-to', None)

        if restricted_to is None:
            minimum = None
            maximum = None
            has_extension_marker = None
        elif isinstance(restricted_to, list):
            if isinstance(restricted_to[0], tuple):
                minimum, maximum = restricted_to[0]
            else:
                minimum = restricted_to[0]
                maximum = restricted_to[0]

            has_extension_marker = (restricted_to[-1] == '...')
        else:
            raise NotImplementedError()

        if isinstance(minimum, str):
            minimum = self.lookup_value(minimum, module_name)[0]['value']

        if isinstance(maximum, str):
            maximum = self.lookup_value(maximum, module_name)[0]['value']

        return minimum, maximum, has_extension_marker

    def is_explicit_tag(self, type_descriptor):
        try:
            return type_descriptor['tag']['kind'] == 'EXPLICIT'
        except KeyError:
            pass

        return False

    def lookup_type_descriptor(self, type_name, module_name):
        module = self._specification[module_name]
        type_descriptor = None

        if type_name in module['types']:
            type_descriptor = module['types'][type_name]
        else:
            for from_module_name, imports in module['imports'].items():
                if type_name in imports:
                    try:
                        from_module = self._specification[from_module_name]
                    except KeyError:
                        raise CompileError(
                            "Module '{}' cannot import type '{}' from missing "
                            "module '{}'.".format(module_name,
                                                  type_name,
                                                  from_module_name))

                    try:
                        type_descriptor = from_module['types'][type_name]
                    except KeyError:
                        raise CompileError(
                            "Type '{}' imported by module '{}' not found "
                            "in module '{}'.".format(type_name,
                                                     module_name,
                                                     from_module_name))

                    module_name = from_module_name
                    break

        if type_descriptor is None:
            raise CompileError("Type '{}' not found in module '{}'.".format(
                type_name,
                module_name))

        return type_descriptor, module_name

    def lookup_value(self, value_name, module_name):
        module = self._specification[module_name]
        value = None

        if value_name in module['values']:
            value = module['values'][value_name]
        else:
            for from_module_name, imports in module['imports'].items():
                if value_name in imports:
                    try:
                        from_module = self._specification[from_module_name]
                    except KeyError:
                        raise CompileError(
                            "Module '{}' cannot import value '{}' from missing "
                            "module '{}'.".format(module_name,
                                                  value_name,
                                                  from_module_name))

                    try:
                        value = from_module['values'][value_name]
                    except KeyError:
                        raise CompileError(
                            "Value '{}' imported by module '{}' not found "
                            "in module '{}'.".format(value_name,
                                                     module_name,
                                                     from_module_name))

                    module_name = from_module_name
                    break

        if value is None:
            raise CompileError("Value '{}' not found in module '{}'.".format(
                value_name,
                module_name))

        return value, module_name

    def lookup_object_class_descriptor(self, object_class_name, module_name):
        module = self._specification[module_name]
        object_class_descriptor = None

        if object_class_name in module['object-classes']:
            object_class_descriptor = module['object-classes'][object_class_name]
        else:
            raise NotImplementedError()

        if object_class_descriptor is None:
            raise CompileError("Object class '{}' not found in module '{}'.".format(
                object_class_name,
                module_name))

        return object_class_descriptor, module_name

    def convert_class_member_type(self, type_descriptor, module_name):
        type_name = type_descriptor['type']
        class_name, member_name = type_name.split('.')
        result = self.lookup_object_class_descriptor(class_name,
                                                     module_name)
        object_class_descriptor, module_name = result
        type_descriptor = deepcopy(type_descriptor)

        for member in object_class_descriptor['members']:
            if member['name'] == member_name:
                type_descriptor['type'] = member['type']
                break

        return type_descriptor


def enum_values_as_dict(values):
    return {
        value[1]: value[0]
        for value in values
        if value != '...'
    }


def enum_values_split(values):
    if '...' in values:
        index = values.index('...')
        return values[:index], values[index + 1:]
    else:
        return values, None


def pre_process(specification):
    return Compiler(specification).pre_process()



EXPECTED = {'CMS-AES-CCM-and-AES-GCM': {'extensibility-implied': False,
                             'imports': {},
                             'object-classes': {},
                             'object-sets': {},
                             'tags': 'IMPLICIT',
                             'types': {'AES-CCM-ICVlen': {'restricted-to': [4,
                                                                            6,
                                                                            8,
                                                                            10,
                                                                            12,
                                                                            14,
                                                                            16],
                                                          'type': 'INTEGER'},
                                       'AES-GCM-ICVlen': {'restricted-to': [12,
                                                                            13,
                                                                            14,
                                                                            15,
                                                                            16],
                                                          'type': 'INTEGER'},
                                       'CCMParameters': {'members': [{'name': 'aes-nonce',
                                                                      'size': [(7,
                                                                                13)],
                                                                      'type': 'OCTET '
                                                                              'STRING'},
                                                                     {'default': 12,
                                                                      'name': 'aes-ICVlen',
                                                                      'type': 'AES-CCM-ICVlen'}],
                                                         'type': 'SEQUENCE'},
                                       'GCMParameters': {'members': [{'name': 'aes-nonce',
                                                                      'type': 'OCTET '
                                                                              'STRING'},
                                                                     {'default': 12,
                                                                      'name': 'aes-ICVlen',
                                                                      'type': 'AES-GCM-ICVlen'}],
                                                         'type': 'SEQUENCE'}},
                             'values': {'aes': {'type': 'OBJECT IDENTIFIER',
                                                'value': [('joint-iso-itu-t',
                                                           2),
                                                          ('country', 16),
                                                          ('us', 840),
                                                          ('organization', 1),
                                                          ('gov', 101),
                                                          ('csor', 3),
                                                          ('nistAlgorithm', 4),
                                                          1]},
                                        'id-aes128-CCM': {'type': 'OBJECT '
                                                                  'IDENTIFIER',
                                                          'value': ['aes', 7]},
                                        'id-aes128-GCM': {'type': 'OBJECT '
                                                                  'IDENTIFIER',
                                                          'value': ['aes', 6]},
                                        'id-aes192-CCM': {'type': 'OBJECT '
                                                                  'IDENTIFIER',
                                                          'value': ['aes', 27]},
                                        'id-aes192-GCM': {'type': 'OBJECT '
                                                                  'IDENTIFIER',
                                                          'value': ['aes', 26]},
                                        'id-aes256-CCM': {'type': 'OBJECT '
                                                                  'IDENTIFIER',
                                                          'value': ['aes', 47]},
                                        'id-aes256-GCM': {'type': 'OBJECT '
                                                                  'IDENTIFIER',
                                                          'value': ['aes',
                                                                    46]}}}}


"""The top level of the ASN.1 tools package contains commonly used
functions and classes, and the command line interface.

"""

import sys
import os
import argparse
import binascii
import logging
import re

from prompt_toolkit.contrib.completers import WordCompleter
from prompt_toolkit import prompt
from prompt_toolkit.history import FileHistory
from prompt_toolkit.interface import AbortAction
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory

from .compiler import compile_dict
from .compiler import compile_string
from .compiler import compile_files
from .compiler import pre_process_dict
from .parser import parse_string
from .parser import parse_files
from .parser import ParseError
from .errors import EncodeError
from .errors import DecodeError
from .errors import CompileError


__author__ = 'Erik Moqvist'
__version__ = '0.62.0'


def _decode_hexstring(codec_spec, asn1_spec, type_name, hexstring):
    try:
        encoded = binascii.unhexlify(hexstring)
    except Exception as e:
        raise TypeError("'{}': {}".format(hexstring, str(e)))

    decoded = codec_spec.decode(type_name, encoded)

    print(asn1_spec.encode(type_name, decoded, indent=4).decode('utf-8'))


def _compile_files(specifications, codec):
    parsed = parse_files(specifications)
    codec_spec = compile_dict(parsed, codec)
    asn1_spec = compile_dict(parsed, 'asn1')

    return codec_spec, asn1_spec


def _do_decode(args):
    codec_spec, asn1_spec = _compile_files(args.specification,
                                           args.codec)

    if args.hexstring == '-':
        for hexstring in sys.stdin:
            hexstring = hexstring.strip('\r\n')

            if hexstring:
                try:
                    _decode_hexstring(codec_spec,
                                      asn1_spec,
                                      args.type,
                                      hexstring)
                except TypeError:
                    print(hexstring)
                except DecodeError as e:
                    print(hexstring)
                    print(str(e))
            else:
                print(hexstring)
    else:
        _decode_hexstring(codec_spec,
                          asn1_spec,
                          args.type,
                          args.hexstring)


def _handle_command_compile(line):
    mo = re.match(r'compile\s+(\w+)\s+(.+)', line)

    if mo:
        try:
            return _compile_files(mo.group(2).split(),
                                  mo.group(1))
        except Exception as e:
            print('error: {}'.format(str(e)))
    else:
        print('Usage: compile <codec> <specification> [<specification> ...]')

    return None, None


def _handle_command_decode(line, codec_spec, asn1_spec):
    if codec_spec:
        mo = re.match(r'decode\s+([^\s]+)\s+(.+)', line)

        if mo:
            hexstring = mo.group(2)

            try:
                _decode_hexstring(codec_spec,
                                  asn1_spec,
                                  mo.group(1),
                                  hexstring)
            except Exception as e:
                print('error: {}'.format(str(e)))
        else:
            print('Usage: decode <type> <hexstring>')
    else:
        print("No compiled specification found. Please use the "
              "'compile' command to compile one.")


def _handle_command_help():
    print('Commands:')
    print('  compile <codec> <specification> [<specification> ...]')
    print('  decode <type> <hexstring>')


def _do_shell(_args):
    commands = ['compile', 'decode', 'help', 'exit']
    completer = WordCompleter(commands, WORD=True)
    user_home = os.path.expanduser('~')
    history = FileHistory(os.path.join(user_home, '.asn1tools-history.txt'))
    codec_spec = None
    asn1_spec = None

    print("\nWelcome to the asn1tools shell.\n")

    while True:
        try:
            line = prompt(u'$ ',
                          completer=completer,
                          complete_while_typing=True,
                          auto_suggest=AutoSuggestFromHistory(),
                          enable_history_search=True,
                          history=history,
                          on_abort=AbortAction.RETRY)
        except EOFError:
            return

        line = line.strip()

        if line:
            if line.startswith('compile'):
                codec_spec, asn1_spec = _handle_command_compile(line)
            elif line.startswith('decode'):
                _handle_command_decode(line, codec_spec, asn1_spec)
            elif line == 'help':
                _handle_command_help()
            elif line == 'exit':
                return
            else:
                print('{}: command not found'.format(line))


def _main():
    parser = argparse.ArgumentParser(
        description='Various ASN.1 utilities.')

    parser.add_argument('-d', '--debug', action='store_true')
    parser.add_argument('-v', '--verbose',
                        choices=(0, 1, 2),
                        type=int,
                        default=1,
                        help=('Control the verbosity; disable(0), warning(1) '
                              'and debug(2). (default: 1).'))
    parser.add_argument('--version',
                        action='version',
                        version=__version__,
                        help='Print version information and exit.')

    # Workaround to make the subparser required in Python 3.
    subparsers = parser.add_subparsers(title='subcommands',
                                       dest='subcommand')
    subparsers.required = True

    # The 'decode' subparser.
    decode_parser = subparsers.add_parser(
        'decode',
        description='Decode given hextring and print it to standard output.')
    decode_parser.add_argument('-c', '--codec',
                               choices=('ber', 'der', 'jer', 'per', 'uper', 'xer'),
                               default='ber',
                               help='Codec (default: ber).')
    decode_parser.add_argument('specification',
                               nargs='+',
                               help='ASN.1 specification as one or more .asn files.')
    decode_parser.add_argument('type', help='Type to decode.')
    decode_parser.add_argument(
        'hexstring',
        help='Hexstring to decode, or - to read hexstrings from standard input.')
    decode_parser.set_defaults(func=_do_decode)

    # The 'shell' subparser.
    shell_parser = subparsers.add_parser('shell',
                                         description='An interactive shell.')
    shell_parser.set_defaults(func=_do_shell)

    args = parser.parse_args()

    levels = [logging.CRITICAL, logging.WARNING, logging.DEBUG]
    level = levels[args.verbose]

    logging.basicConfig(format='%(levelname)s: %(message)s',
                        level=level)

    if args.debug:
        args.func(args)
    else:
        try:
            args.func(args)
        except BaseException as e:
            sys.exit('error: {}'.format(str(e)))
